/*  htshd.c

    Hypertext shell daemon

    Copyright, Exolution GmbH / Michael Kerrisk, Sep 2000
    http://www.exolution.com/wapsh   mailto:wapsh@exolution.de 

    This software is Open Source licensed under the terms of the Trolltech QPL 
    version 1.0.  See http://www.exolution.de/wapsh for further information.
    

    This is the server daemon for the htsh application, which allows users 
    to login to a shell via a web browser or WAP phone.  This daemon creates
    multiple processes to handle user logins, as follows:

    Master Command Server (MCS)
    	This is the overall server process which accepts incoming login 
	requests (via a socket connected to a well-known port) 
	and creates a Child Command Server to handle each login.
	If htshd is started via inetd(8), then no MCS process as such is 
	created.  Instead, each htshd process instantiated by inetd(8)
	automatically configures itself as a CCS process.

    Child Command Server (CCS)
    	Each login creates a CCS process, which in turn creates a 
	corresponding Child Shell (CSh) process.  After initially 
	authenticating the user, the CCS simply acts as a conduit, 
	transferring data in both directions between the HTTP Server 
	(via a socket) and the Child Shell (via a pseudo-terminal).  

    Child Shell (CSh)
    	This process runs the login shell for the user.  It is connected
	to the CCS process via a pseudo-terminal.

    So far successfully ported to:

    	Linux 2.2 (2.2.14, glibc 2, Suse 6.4), Linux 2.4.0test
    	FreeBSD 4.1.1
    	SunOS 5.7
*/


#ifdef __sun__

char *crypt(const char *key, const char *salt);                            

#else

/* Following are to get crypt() defn - but this breaks things on Sun */

#define _XOPEN_SOURCE
#define _GNU_SOURCE

#endif

#ifdef __NetBSD__
#undef _XOPEN_SOURCE
#endif


#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>

#include <ctype.h>
#include <errno.h>	
#include <fcntl.h>
#include <fnmatch.h>
#include <grp.h>
#include <limits.h>
#include <netdb.h>
#include <pwd.h>
#ifdef USE_SHADOW_PWD
#include <shadow.h>
#endif
#include <signal.h>
#include <stdarg.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <syslog.h>
#include <termios.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/utsname.h>



#ifndef INADDR_NONE		/* Some Unices don't define this */
#define INADDR_NONE -1
#endif

#define SERVICE_NAME "htsh"	/* Should be in /etc/services
                                   on login and http server hosts */

#define CONFIG_FILE "/etc/htshd.conf"
				/* Default name of startup config file */

#define GLOBAL_RC_FILE "/etc/htsh_profile"
				/* Default name of global user 
				   initialisation file */
#define RCFILE ".htshrc"	/* Name of user-specific initialisation file
                                   in user's home directory */


	/***** MCS exit status values *****/

#define EXIT_BAD_CMD_LINE	2	/* Exit status if bad command line 
					   options specified */

typedef enum { FALSE, TRUE } Boolean;	

#define min(m,n) ((m) < (n) ? (m) : (n))
#define max(m,n) ((m) > (n) ? (m) : (n))


	/***** Default values for config file / cmd line options settings *****/

const int DEF_INPUT_TIMEOUT = 1800;	
                                /* Def timeout for shell input (secs) */
const float DEF_OUTPUT_TIMEOUT = 0.5;	
                                /* Def timeout for shell output (secs) */
const float MIN_OUTPUT_TIMEOUT = 0.1; 	
                                /* Min timeout for shell output (secs) */
const float MAX_OUTPUT_TIMEOUT = 15.0;	
                                /* Max timeout for shell output (secs) */
const int DEF_CS_MAX_TRANSFER_SIZE = 10000;
                                /* Default maximum transfer size between Child 
                                   Shell and HTTP Server (bytes) */
const int DEF_OUTPUT_BUFFER_LIMIT = 100000;
				/* Default size for per-user output buffers */

const int SHORTCUT_BLOCK_SIZE_LOW_BOUND = 3;
				/* Lower bound for shortcutblocksize */

const int HISTORY_BLOCK_SIZE_LOW_BOUND = 3;
				/* Lower bound for historyblocksize */

const int CS_MAX_TRANSFER_SIZE_LOW_BOUND = 1000;
				/* Lower bound for csmaxtransfersize */

const int OUTPUT_WINDOW_SIZE_LOW_BOUND = 100;
				/* Lower bound for outputwindowsize*/

	/****** Message types from HTTP server to CCS ******/

#define CMD_SHELLCMD 	'c'	/* Send text input to shell */
#define CMD_CONTROL 	'^'	/* Send a control character to shell */
#define CMD_SHORTCUT 	's'	/* Send one of the shortcut commands to shell */
#define CMD_SIGNAL 	'x'	/* Send a signal to shell */
#define CMD_CLOSE 	'z'	/* Close CCS and Child Shell */
#define CMD_NULL 	'n'	/* Null shell input (used to send outstanding
                                   shell output back to client) */

	/****** Message types from CCS to HTTP Server ******/

#define RESP_NORMAL	"n"	/* Normal: shell output follows */
#define RESP_ERROR	"e"	/* Error: message follows, shell
                                   session is still alive */
#define RESP_FATAL	"f"	/* Fatal: shell session has died,
                                   user should re-login */


	/***** command line option and config file settings *****/

/* List of users permitted to login.  If userSetSize is 0,
   then any user may log in */

#define MAX_USER_SET_SIZE 1000
char *userSet[MAX_USER_SET_SIZE]; /* Array of permitted usernames */
int userSetSize = 0;		/* Number of used elements in userSet[] */


/* List of HTTP servers permitted to start shell sessions on this 
   login server.  If numHttpServers = 0, then logins are only accepted from
   an HTTP Server on localhost */

#define MAX_HTTP_SERVERS 100
int numHttpServers = 0;		
int httpServer[MAX_HTTP_SERVERS];


/* Set of protocols which are allowed to login. */
  
#define PRNAME_WAP "wap"
#define PRNAME_HTTP "http"

	/* Octal protocol masks */

#define PROTO_HTTP 01
#define PROTO_WAP  02

int allowedProtocols = PROTO_WAP | PROTO_HTTP;


char *service;			/* The service/port number at which we 
				   should listen */
Boolean ignoreStdConfigFile;	/* Process config file? */
int debugMask;			/* Controls what debugging output appears */
Boolean runAsDaemon;		/* Should we run as a daemon (if started 
				   manually)? */
Boolean weAreDaemon;		/* Are we running as a daemon (used to 
				   control where debug and error messages 
				   are sent (i.e. stderr or syslog)) */
int shellTimeout;		/* Shell timeout */	
int csMaxTransferSize;		/* Maximum single block of bytes to transfer 
				   between CCS and HTTP Server */
float csOutputTimeout;		/* Shell output timeout */
char *globalRcFile;		/* Name of startup initialisation file to be 
				   used by all users. */
int outputBufferLimit;		/* Upper limit for size of output buffer */
char *logFileDir = NULL;	/* Directory into which to write session 
				   log files */



	/***** CCS/CSh global vars *****/

char ptyName[11];		/* Name of pty slave to be used by CSh */



/* Return true if given file descriptor is a socket */

Boolean
isSocket(int fd)
{
    struct stat sb;

    if (fstat(fd, &sb) == -1) syslog(LOG_ERR, "isSocket: fstat");
    return S_ISSOCK(sb.st_mode);
} /* isSocket */



	/******************************************/
	/**** Debug and error report routines *****/
	/******************************************/

void
rawErrorMsg(const char *format, va_list argList)
{
    if (weAreDaemon || isSocket(STDERR_FILENO)) {	
		
	char buf[1000];

	vsnprintf(buf, sizeof(buf), format, argList);
    	syslog(LOG_ERR, buf);
    } else {
        vfprintf(stderr, format, argList);
	fprintf(stderr, "\n");
    } /* if */
} /* rawErrorMsg */



/* Write an error message to stderr or the system log, depending on 
   whether we are a daemon */

void
errorMsg(const char *format, ...)
{
    va_list argList;
    
    va_start(argList, format);

    rawErrorMsg(format, argList);
    
    va_end(argList);
} /* errorMsg */


/* Write a debug message to stderr or the system log, depending on whether 
   we are a daemon, and also depending on (dMask & debugMask) setting */

#define DBG_MCS 04000
#define DBG_CCS 02000
#define DBG_CSH 01000

#define DBG_OUT 00100

#define DBG_RCF 00004
#define DBG_CFG 00001

void
dbgMsg(int dMask, const char *format, ...)
{
    va_list argList;

    if (dMask == 0 || (dMask & debugMask)) {
        va_start(argList, format);

	/*
	fprintf(stderr, "DBGMSG weAreDaemon=%d, isSock=%d\n", weAreDaemon,
		isSocket(STDERR_FILENO));
	*/
	if (weAreDaemon || isSocket(STDERR_FILENO)) {

	    char buf[1000];
	   
	    vsnprintf(buf, sizeof(buf), format, argList);
    	    syslog(LOG_DEBUG, buf);
	} else {
            vfprintf(stderr, format, argList);
            fprintf(stderr, "\n");
	} /* if */
	
        va_end(argList);
    } /* if */
} /* dbgMsg */



int errCoreDump = 0;            /* Do a core dump on error or just exit()? */


static void
displayErrnoString(void)
{
    errorMsg("System call error: %d (%s), occurred during...", 
		errno, strerror(errno));
} /* displayErrnoString */



/* Diagnose a system call error */

void
sysErrMsg(const char *format, ...)
{
    va_list    	argList;

    displayErrnoString();

    va_start(argList, format);
    rawErrorMsg(format, argList);
    va_end(argList);
} /* sysErrMsg */



/* Diagnose a system call error and terminate the process */

void
sysErrExit(const char *format, ...)
{
    va_list    	argList;

    displayErrnoString();

    va_start(argList, format);
    rawErrorMsg(format, argList);
    errorMsg("Exiting after system call error");
    va_end(argList);

    if (errCoreDump) 		/* Produces a core dump for debugger */
        abort();
    else
	exit(EXIT_FAILURE);
} /* sysErrExit */



/* Print an error message and terminate the process */

void
fatalErr(const char *format, ...)
{
    va_list        argList;

    fflush(stdout);		/* First flush any pending stdout */

    va_start(argList, format);
    rawErrorMsg(format, argList);
    errorMsg("Exiting after fatal error");
    va_end(argList);

    if (errCoreDump) 		/* Produces a core dump for debugger */
        abort();
    else
	exit(EXIT_FAILURE);
} /* fatalErr */



#ifndef HAVE_SETENV       /* SunOS (5.7) doesn't have setenv */

void
nunsetenv(const char *name)
{
    extern char **environ;
    char **ep;
    int len;

    len = strlen(name);

    for (ep = environ; *ep != NULL; ep++) 
	if (strncmp(*ep, name, len) == 0 && (*ep)[len] == '=')
	    *ep = NULL;
} /* nunsetenv */



int
setenv(const char *name, const char *value, int overwrite)
{
    char *es;
    int s;
   
    dbgMsg(DBG_CSH | DBG_CSH | DBG_MCS, "setenv: %s %s", name, value);
    if (getenv(name) != NULL && overwrite == 0) return 1;

    nunsetenv(name);	/* Remove all occurrences */

    es = malloc(strlen(name) + strlen(value) + 2);
    			/* +2 for '=' and null terminator */
    if (es == NULL) return -1;
    
    strcpy(es, name);
    strcat(es, "=");
    strcat(es, value);

    s = putenv(es);

    return s;
} /* setenv */

#endif



	/******************************************/
	/***** Miscellaneous utility routines *****/
	/******************************************/

/* Return true if given string is a floating point number, which
   may or may not contain a decimal point */

Boolean
isFloatNumber(const char *s)
{
    int dotCount;

    dotCount = 0;
    
    while (*s != '\0') {
        if (!(*s == '.' || isdigit((int) *s))) {
            return FALSE;
	} else {
	    if (*s == '.') {
	    	dotCount++;
		if (dotCount > 1) return FALSE;
	    } /* if */
	    s++;
        } /* if */
    } /* while */

    return TRUE;
} /* isFloatNumber */



/* Return true if given string is an integr number in given base */

Boolean
isIntNumber(const char *s, int base)
{
    if (base > 10)
	fatalErr("Bad base (%d) given to isIntNum", base);
    
    while (*s != '\0') {
        if (!isdigit((int) *s) || (*s - '0') >= base)
            return FALSE;
        else 
	    s++;
    } /* while */
    return TRUE;
} /* isIntNumber */



	/**********************************/
	/***** utmp and wtmp routines *****/
	/**********************************/


#if defined (__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define USE_LOGIN
#ifdef __FreeBSD__
#include <libutil.h>
#else
#include <util.h>
#endif
#endif

#if defined(__linux__) || defined(__sun__)
#define USE_EXTENDED_UTMP
#endif


#ifdef __sun__
#include <utmpx.h>
#define ut_time ut_tv.tv_sec
#define updwtmp updwtmpx
#define pututline pututxline
#define setutent setutxent
#define endutent endutxent

#ifndef UT_HOSTSIZE
#define UT_HOSTSIZE 257
#endif

#ifndef _PATH_WTMP
#define _PATH_WTMP "/var/adm/wtmp"
#endif

#include <lastlog.h>
#ifndef _PATH_LASTLOG
#define _PATH_LASTLOG "/var/adm/lastlog"
#endif

#else

#include <utmp.h>

#endif


char remoteHost[UT_HOSTSIZE];	/* DNS / IP address of remote server */

/* Create entries in the utmp and wtmp files for this login, 
   and update the lastlog entry as well */

void
createUtmpAndWtmpEntries(pid_t pid, struct passwd *pwd, char *ttyName)
{
#ifdef __sun__
    static struct utmpx utmpEnt;
#else
    static struct utmp utmpEnt;
#endif
    struct lastlog lastlogEnt;
    int fd;
    
    memset(&utmpEnt, 0, sizeof(utmpEnt));

#ifdef USE_EXTENDED_UTMP
    utmpEnt.ut_type = USER_PROCESS;
    utmpEnt.ut_pid = pid;
#endif
    
    strcpy(utmpEnt.ut_line, ttyName + strlen("/dev/"));
    /* only correct for ptys named /dev/tty[pqr][0-9a-z] */
#ifdef USE_EXTENDED_UTMP
    strcpy(utmpEnt.ut_id, ttyName + strlen("/dev/tty"));
#endif

    strncpy(utmpEnt.ut_host, remoteHost, UT_HOSTSIZE);

    time(&utmpEnt.ut_time);
    strcpy(utmpEnt.ut_name, pwd->pw_name);
    

#ifdef USE_LOGIN
    dbgMsg(DBG_CSH | DBG_CCS, "Login: /%s/ - %s, %s", 
	    utmpEnt.ut_line, utmpEnt.ut_name, utmpEnt.ut_host);
    login(&utmpEnt);
#else
    dbgMsg(DBG_CCS | DBG_CSH, 
	    "Creating utmp entry for %d, %s, line=%s"
#ifdef USE_EXTENDED_UTMP
	    ", id=%s"
#endif
	    , 
	    pid, pwd->pw_name, utmpEnt.ut_line
#ifdef USE_EXTENDED_UTMP
	    , utmpEnt.ut_id
#endif
	    );
    setutent();
    pututline(&utmpEnt);
    updwtmp(_PATH_WTMP, &utmpEnt);
    endutent();

#endif

    /* Create a lastlog entry */
    
    dbgMsg(DBG_CCS | DBG_CSH, "Create lastlog entry, tty=%s, host=%s", 
	    ttyName, remoteHost);
    fd = open(_PATH_LASTLOG, O_WRONLY);

    if (fd == -1) {
	sysErrMsg("Couldn't open: %s", _PATH_LASTLOG);
    } else {
	dbgMsg(DBG_CCS | DBG_CSH, "lseek to %d*%d in %d - %s", pwd->pw_uid, 
		sizeof(struct lastlog), fd, _PATH_LASTLOG);

        if (lseek(fd, pwd->pw_uid * sizeof(struct lastlog), SEEK_SET) == -1) {
	    sysErrMsg("lseek lastlog");
	} else {
            strcpy(lastlogEnt.ll_line, ttyName + strlen("/dev/"));
            time(&lastlogEnt.ll_time);
            strncpy(lastlogEnt.ll_host, remoteHost, sizeof(lastlogEnt.ll_host));
	    if (write(fd, &lastlogEnt, sizeof(lastlogEnt)) == -1) 
	        sysErrMsg("write lastlog: %s %d", _PATH_LASTLOG, fd);
	} /* if */
        close(fd);
    } /* if */ 
} /* createUtmpAndWtmpEntries */



/* Clear utmp and wtmp entries for this login */

void
clearUtmpAndWtmpEntries(pid_t pid)
{
#ifdef __sun__
    struct utmpx utmpEnt;
#else
    struct utmp utmpEnt;
#endif
#ifdef USE_LOGIN
    int s;
#endif
 
    memset(&utmpEnt, 0, sizeof(utmpEnt));

    time(&utmpEnt.ut_time);

#ifdef USE_EXTENDED_UTMP
    utmpEnt.ut_type = DEAD_PROCESS;
    utmpEnt.ut_pid = pid;
#endif
    
    strcpy(utmpEnt.ut_line, ptyName + strlen("/dev/"));
    /* only correct for ptys named /dev/tty[pqr][0-9a-z] */
#ifdef USE_EXTENDED_UTMP
    strcpy(utmpEnt.ut_id, ptyName + strlen("/dev/tty"));
#endif

    dbgMsg(DBG_CCS | DBG_CSH, "Clearing utmp entry for PID %d, line=%s"
#ifdef USE_EXTENDED_UTMP
	    ", id=%s\n"
#endif
	    , 
	    pid, utmpEnt.ut_line
#ifdef USE_EXTENDED_UTMP
	    , utmpEnt.ut_id
#endif
    );

    dbgMsg(DBG_CCS | DBG_CSH, 
	    "Clearing wtmp entry for line=%s"
#ifdef USE_EXTENDED_UTMP
	    ", pid=%d. id=%s type=%d\n" 
#endif
	    ,
	    utmpEnt.ut_line
#ifdef USE_EXTENDED_UTMP
	    , utmpEnt.ut_pid, utmpEnt.ut_id, utmpEnt.ut_type
#endif
	    );
#ifdef USE_LOGIN
    s = logout(utmpEnt.ut_line);
    dbgMsg(DBG_CCS | DBG_CCS, "Logout: /%s/ returned %d", utmpEnt.ut_line, s);
    logwtmp(utmpEnt.ut_line, "", "");
#else
    setutent();
    pututline(&utmpEnt);
    updwtmp(_PATH_WTMP, &utmpEnt);
    endutent();
#endif
} /* clearUtmpAndWtmpEntries */



        /*************************************/
        /***** Pseudo-terminal routines ******/
        /*************************************/

/* Here we use BSD style ptys.  Code largely based on WR Stevens APUE */

/* Open a pseudo-terminal master, return the file descriptor, or -1 on error.
   The name of the corresponding pseudo-terminal slave is returned in ptyName.
*/

int
ptyMasterOpen(char *ptyName)
{
    int ptyMasterFd;
    char *p1, *p2;

    /* Open pty master the BSD way... */

                   /*0123456789*/
    strcpy(ptyName, "/dev/ptyXX");
    for (p1 = "pqrstuvwxyzabcde"; *p1 != '\0'; p1++) {
    	ptyName[8] = *p1;
        for (p2 = "0123456789abcdef"; *p2 != '\0'; p2++) {
    	    ptyName[9] = *p2;
            ptyMasterFd = open(ptyName, O_RDWR);

            if (ptyMasterFd == -1) {
            	if (errno == EIO) {		/* pty already in use */
                    continue;		
                } else {	/* Some other error (e.g. no such pty name) */
            	    sysErrMsg("ptyMasterOpen: %s", ptyName);
                    return -1;
                } /* if */
            } else {

            	/* Succesfully opened master: Return name of slave ("tty")
                   corresponding to master ("pty") */

            	ptyName[5] = 't';
		if (chmod(ptyName, 0666) == -1) 
		    sysErrExit("ptyMasterOpen: chmod - %s", ptyName);

                return ptyMasterFd;
            } /* if */
        } /* for */
    } /* for */

    return -1;		/* Didn't find a pty after trying all possibilities */
} /* ptyMasterOpen */



/* Open a pseudo-terminal slave, returning a file descriptor, or -1 on error */

int
ptySlaveOpen(const char *ptyName)
{
    int ptySlaveFd;

    dbgMsg(DBG_CSH, "ptySlaveOpen: %s", ptyName);
    ptySlaveFd = open(ptyName, O_RDWR);
    if (ptySlaveFd == -1) sysErrMsg("ptySlaveOpen");
    return ptySlaveFd;
} /* ptySlaveOpen */



/* Set the ownership and permissions of the pty slave.

   The UID of the pty is set to be the UID of the logged in user,
   and the GID is set to the group "tty" (or the GID of the logged in user]
   if the "tty" group does not exist 

   Returns 0 on success or -1 on error */

uid_t prevPtySlaveUid;
gid_t prevPtySlaveGid;
mode_t prevPtySlaveMode;

int
setPtyOwnerAndMode(const char *ptyName, const struct passwd *pwd)
{
    struct stat sb;
    struct group *grp;
    int NEW_MODE = 0620;
    gid_t newGid;

    if (stat(ptyName, &sb) == -1) {
	sysErrMsg("setPtyOwnerAndMode: stat %s", ptyName);
	return -1;
    } /* if */

    prevPtySlaveUid = sb.st_uid;
    prevPtySlaveGid = sb.st_gid;
    prevPtySlaveMode = sb.st_mode;

    dbgMsg(DBG_CSH, "setPtyOwnerAndMode: previous: UID=%d, GID=%d, mode=%o",
	    prevPtySlaveUid, prevPtySlaveGid, prevPtySlaveMode);

    grp = getgrnam("tty");
    newGid = (grp != NULL) ? grp->gr_gid : pwd->pw_gid;

    dbgMsg(DBG_CSH, "setPtyOwnerAndMode: setting: UID=%d, GID=%d, mode=%o",
	    pwd->pw_uid, newGid, NEW_MODE);
    
    if (chown(ptyName, pwd->pw_uid, newGid) == -1) {
	sysErrMsg("setPtyOwnerAndMode: chown - %s", ptyName);
	return -1;
    } /* if */

    if (chmod(ptyName, NEW_MODE) == -1) {
	sysErrMsg("setPtyOwnerAndMode: chmod - %s", ptyName);
	return -1;
    } /* if */

    return 0;
} /* setPtyOwnerAndMode */



/* Reset pty slave ownership and perms to the values saved by a previous
   call to setPtyOwnerAndMode */

int
resetPtyOwnerAndMode(char *ptyName)
{
    dbgMsg(DBG_CSH, "resetPtyOwnerAndMode: setting: UID=%d, GID=%d, mode=%o",
	    prevPtySlaveUid, prevPtySlaveGid, prevPtySlaveMode);
    
    if (chown(ptyName, prevPtySlaveUid, prevPtySlaveGid) == -1) {
	sysErrMsg("resetPtyOwnerAndMode: chown - %s", ptyName);
	return -1;
    } /* if */

    if (chmod(ptyName, prevPtySlaveMode) == -1) {
	sysErrMsg("resetPtyOwnerAndMode: chmod - %s", ptyName);
	return -1;
    } /* if */

    return 0;
} /* resetPtyOwnerAndMode */



	/****************************/
	/***** SIGCHLD Handlers *****/
	/****************************/

/* SIGCHLD catcher used by MCS to reap dead CCS children */

void
ccsReaper(int signum)
{
    int status;
    pid_t pid;

    while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
        dbgMsg(DBG_MCS | DBG_CCS, "CCS %d terminated by %s %d", pid,
                WIFEXITED(status) ? "exit" : "signal",
                WIFEXITED(status) ? WEXITSTATUS(status) : WTERMSIG(status));
    } /* while */
} /* ccsReaper */



/* SIGCHLD catcher used by CCS to reap dead Child Shell children */

Boolean childShellAlive;

void
childShellReaper(int signum)
{
    int status;
    pid_t pid;

    while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
        dbgMsg(DBG_CCS | DBG_CCS, "Child Shell %d terminated by %s %d", pid,
                WIFEXITED(status) ? "exit" : "signal",
                WIFEXITED(status) ? WEXITSTATUS(status) : WTERMSIG(status));
        childShellAlive = FALSE;	/* Each CCS has only one Child Shell */
	
	clearUtmpAndWtmpEntries(pid);

	(void) resetPtyOwnerAndMode(ptyName);
    } /* while */
} /* childShellReaper */


	/***************************/
	/***** Socket Routines *****/
	/***************************/

/* Build (in inetAddr) the internet address for given
   host, service and socketType.  Return TRUE if successful */

Boolean
buildInetAddress(const char *host, const char *service, int socketType, 
	struct sockaddr_in *inetAddr)
{
     struct servent *serviceInfo;
     struct hostent *hostInfo;
     char *protocol;
     Boolean isNumericService;
     int j;

    /* clear address structure */

    memset(inetAddr, 0, sizeof(struct sockaddr_in));

    /* sin_family is always AF_INET for Internet domain */

    inetAddr->sin_family = AF_INET;


    /***** Generate the host address ******/

    if (strlen(host) == 0) {

	/* Blank host string ==> use the wildcard host address.  This is 
	   useful for servers that have multiple INTERNET addresses: 
	   by binding to the wildcard address, the server can accept 
	   connection requests on any of its host addresses */

	inetAddr->sin_addr.s_addr = INADDR_ANY;
    
    } else if (strcmp(host, "*") == 0) {

        /* "*" ==> datagram broadcast address: can be used to
	   broadcast a message to multiple servers listening
           at the same port # on different hosts */

	inetAddr->sin_addr.s_addr = INADDR_BROADCAST;

    } else {

	/* First try converting the host name as though it is an
	   Internet dot notation address (a.b.c.d) */

        if ((inetAddr->sin_addr.s_addr = inet_addr(host)) == INADDR_NONE) {

            /* If that fails try looking up address by host name */

            if ((hostInfo = gethostbyname(host)) == NULL) {
                errorMsg("%s not found in /etc/hosts", host);
                return FALSE;
            } /* if */

	    memcpy(&(inetAddr->sin_addr), hostInfo->h_addr, hostInfo->h_length);
	} /* if */
    } /* if */


    /***** Generate the port number *****/

    /* If the service name is numeric, just convert it to an
       integer otherwise look up the name in the service file */

    isNumericService = TRUE;
    for (j = 0; j < strlen(service) && isNumericService; j++)
	isNumericService = isdigit((int) service[j]); 

    if (isNumericService) {
        inetAddr->sin_port = htons(atoi(service));
    } else {
        if (socketType == SOCK_STREAM) {
	    protocol = "tcp";
        } else if (socketType == SOCK_DGRAM) {
	    protocol = "udp";
        } else {
	    errorMsg("buildInetAddress: Bad socket type");
            return -1;
        } /* if */

        serviceInfo = getservbyname(service, protocol);
        if (serviceInfo == NULL) {
            errorMsg("Service '%s' (protocol '%s') not found in /etc/services", 
		    service, protocol);
            return FALSE;
        } /* if */

        inetAddr->sin_port = serviceInfo->s_port;
    } /* if */

    return TRUE;
} /* buildInetAddress */



/* Return pointer to string containing socket address in
   ASCII form: "(host, port#)" */

char *
inetAddressString(const struct sockaddr_in *inetAddr)
{
    static char buf[100];
    struct hostent *hostInfo;
    char *hostname;


    hostInfo = gethostbyaddr((char *) &inetAddr->sin_addr,
		sizeof(inetAddr->sin_addr), inetAddr->sin_family);

    if (hostInfo == NULL)

        /* If the host information is unavailable, just format its
           Internet address in the "Internet dot format". */

        hostname = inet_ntoa(inetAddr->sin_addr);
    else
        hostname = hostInfo->h_name;  /* point to host's name */

    sprintf(buf, "(%s, %d)", hostname, ntohs(inetAddr->sin_port));
			/* Convert port# to host byte order */
    return buf;
} /* inetAddressString */


char *
socketAddressString(int fd)
{
    struct sockaddr_in sockAddr;
    int addrLen;

    addrLen = sizeof(sockAddr);
    if (getsockname(fd, (struct sockaddr *) &sockAddr, &addrLen) == -1) {
    	sysErrMsg("socketAddressString: getsockname");
    	return NULL;
    } /* if */

    return inetAddressString(&sockAddr);
} /* socketAddressString */



/* Create an INTERNET domain socket and bind it to the address
   { wildcard-host-address, service, socketType/protocol }.

   Return the socket descriptor on success, or -1 on error */

static int
inetSocket(const char *service, int socketType)
{
    int sd;
    struct sockaddr_in socketAddress;

    if (!buildInetAddress("", service, socketType, &socketAddress)) {
        errorMsg("inetSocket: Can't build socket address");
	return -1;
    } /* if */

    sd = socket (AF_INET, socketType, 0);
    if (sd == -1) {
	sysErrMsg("inetSocket: socket");
	return -1;
    } /* if */

    if (socketType == SOCK_STREAM) {
    	int ov = 1;

        if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &ov, sizeof(ov)) == -1) {
	    sysErrMsg("setsockopt - SO_REUSEADDR");
	    return -1;
    	} /* if */
    } /* if */

    if (bind(sd, (struct sockaddr *) &socketAddress, 
		sizeof(socketAddress)) == -1) {
        sysErrMsg("inetSocket: bind");
	close(sd);
	return -1;
    } /* if */

    if (socketType == SOCK_STREAM) {
        if (listen(sd, 1) == -1) {
            sysErrMsg("inetSocket: listen");
	    close(sd);
	    return -1;
        } /* if */
    } /* if */

    return sd;
} /* inetSocket */


/* Create an INTERNET domain TCP socket */

int
inetStreamSocket(const char *service)
{
    return inetSocket(service, SOCK_STREAM);
} /* inetStreamSocket */


/* Create an INTERNET domain UDP socket */

int
inetDatagramSocket(const char *service)
{
    return inetSocket(service, SOCK_DGRAM);
} /* inetDatagramSocket */



/* Create an INTERNET domain socket and connect it to the
   remote address specified by the parameters.

   Return the socket descriptor on success, or -1 on error */

int
inetConnect(const char *host, const char *service, int socketType)
{
    int sd;
    struct sockaddr_in socketAddress;

    if (!buildInetAddress(host, service, socketType, &socketAddress)) {
        errorMsg("Couldn't build socket address");
 	return -1;
    } /* if */

    sd = socket(AF_INET, socketType, 0);
    if (sd == -1) {
	sysErrMsg("inetConnect: socket");
	return -1;
    } /* if */

    if (connect(sd, (struct sockaddr *) &socketAddress, 
		sizeof(socketAddress)) == -1) {
        sysErrMsg("inetConnect: connect");
        close(sd);
 	return -1;
    } /* if */

    return sd;
} /* inetConnect */




        /*****************************************************************/
        /***** Data structures and routines for initialisation files *****/
        /*****************************************************************/

/* RCFILE 'set -o/+o' options and corresponding bitmask values */

#define OPT_ALLOWSHELLCMD	 	 01
#define OPTNAME_ALLOWSHELLCMD 		"allowshellcmd"

#define OPT_DISPLAYMENU		 	 02
#define OPTNAME_DISPLAYMENU 		"displaymenu"

#define OPT_ALLOWSILENT		 	 04
#define OPTNAME_ALLOWSILENT	 	"allowsilent"

#define OPT_ALLOWCONTROLCHARS	 	010
#define OPTNAME_ALLOWCONTROLCHARS 	"allowcontrolchars"

#define OPT_HISTORY 			020
#define OPTNAME_HISTORY		 	"history"

#define OPT_FILTERANSIESC 		040
#define OPTNAME_FILTERANSIESC	 	"filteransiesc"

#define OPT_ALLOWTRIGRAPHS 		0100
#define OPTNAME_ALLOWTRIGRAPHS	 	"allowtrigraphs"

#define OPT_ALLOWUSERINIT 		010000
#define OPTNAME_ALLOWUSERINIT	 	"allowuserinit"

#define OPT_SESSIONLOGGING		0100000

/* Bitmasks for options to the RCFILE "sc" (shortcut) command */

#define SCOPT_NONL 	01		/* Don't append newline to 
					   shortcut string */


#define MAX_RCFILE_SHORTCUTS 200	/* Maximum number of 'sc' commands in
                                           initialisation file */

#define MAX_RCFILE_SET_VALUES 50	/* Maximum number of 'set NAME VALUE'
                                           commands in initialisation file */

#define MAX_RCFILE_ERRORS 10		/* Maximum number of errors in
                                           initialisation file before we
                                           give up */

/* Parsed information from RCFILE */

struct RcInfo {
    int setOptions;			/* Options set by 'set' command */
    int numSc;				/* Total number of shortcuts defined */
    int scOptions[MAX_RCFILE_SHORTCUTS];/* List of shortcut options */
    char *scName[MAX_RCFILE_SHORTCUTS];	/* List of shortcut names */
    char *scDef[MAX_RCFILE_SHORTCUTS];	/* List of shortcut definitions */
    int numSetVals;			/* Total number of 'set NAME VALUE' 
                                           commands */
    char *setName[MAX_RCFILE_SET_VALUES];/* Name from 'set NAME VALUE' */
    char *setVal[MAX_RCFILE_SET_VALUES];/* Value from 'set NAME VALUE' */
    int numErrors;			/* Number of errors during parsing */
    int errorLine[MAX_RCFILE_ERRORS];	/* List of lines which generated parse
    					   errors */
    char *errorMsg[MAX_RCFILE_ERRORS];	/* Message text */
};


struct RcInfo rcInfo;



/* Initialise an RcInfo structure */

void
initRcInfo(struct RcInfo *rcInfo)
{
    rcInfo->numSc = 0;
    rcInfo->numErrors = 0;
    rcInfo->numSetVals = 0;
    rcInfo->setOptions = 0;
} /* initRcInfo */



/* Free heap memory associated with RcInfo structure, and initialise
   the structure */

void
freeRcInfo(struct RcInfo *rcInfo)
{
    int j;

    for (j = 0; j < rcInfo->numSc; j++) {
        free(rcInfo->scName[j]);
        free(rcInfo->scDef[j]);
    } /* for */

    initRcInfo(rcInfo);
} /* freeRcInfo */



/* Display parsed initialisation file information */

void
printRcInfo(const struct RcInfo *rcInfo)
{
    int j;

    dbgMsg(DBG_RCF, "-------------------- rcInfo --------------------");
    dbgMsg(DBG_RCF, "setOptions=%o (oct)", rcInfo->setOptions);
    dbgMsg(DBG_RCF, "numSc=%d", rcInfo->numSc);
    for (j = 0; j < rcInfo->numSc; j++)
        dbgMsg(DBG_RCF, "sc %d: //%s// is // %s// Options=%o", j, 
		rcInfo->scName[j], rcInfo->scDef[j], rcInfo->scOptions[j]);
    dbgMsg(DBG_RCF, "numErrors=%d", rcInfo->numErrors);
    if (rcInfo->numErrors > 0) {
    	dbgMsg(DBG_RCF, "Errors in lines: ");
        for (j = 0; j < rcInfo->numErrors; j++)
            dbgMsg(DBG_RCF, "%d: %s", rcInfo->errorLine[j], 
		    rcInfo->errorMsg[j]);
    } /* if */
    dbgMsg(DBG_RCF, "------------------ end rcInfo ------------------");
} /* printRcInfo */



        /*****************************************************/
        /***** readn() and writen() from WR Stevens APUE *****/
        /*****************************************************/


ssize_t			/* Read "n" bytes from a descriptor. */
readn(int fd, void *vptr, size_t n)
{
    size_t	nleft, nread;
    char	*ptr;

    ptr = vptr;
    nleft = n;
    while (nleft > 0) {
        if ((nread = read(fd, ptr, nleft)) < 0)
            return(nread);			/* error, return < 0 */
        else if (nread == 0)
            break;				/* EOF */

        nleft -= nread;
        ptr   += nread;
    } /* while */
    return(n - nleft);		/* return >= 0 */
} /* readn */



ssize_t			/* Write "n" bytes to a descriptor. */
writen(int fd, const void *vptr, size_t n)
{
    size_t	nleft, nwritten;
    const char	*ptr;

    ptr = vptr;	/* can't do pointer arithmetic on void* */
    nleft = n;
    while (nleft > 0) {
        if ((nwritten = write(fd, ptr, nleft)) <= 0)
            return(nwritten);		/* error */

        nleft -= nwritten;
        ptr   += nwritten;
    } /* while */
    return(n);
} /* writen */



/* Print formatted output to a file descriptor */

void
fdPrint(int fd, const char *format, ...)
{
    va_list argList;
    const int FDBUF_SIZE = 10000;
    char buf[FDBUF_SIZE];
    int len, numSent;

    va_start(argList, format);
    len = vsnprintf(buf, FDBUF_SIZE, format, argList);

    if (len >= FDBUF_SIZE)
	syslog(LOG_WARNING, "fdPrint: Warning large buffer size in fdPrint");

    numSent = writen(fd, buf, len+1);

    if (numSent != len + 1)
        fatalErr("fdPrint: send failed: %d errno=%d", numSent, errno);
    va_end(argList);
} /* fdPrint */



        /**************************************************/
        /****** Routines to parse initialisation file *****/
        /**************************************************/

/* Scan a string containing space separated protocol names and use
   them to build a corresdponding protocol bitmask

   Returns a bitmask of the protocols, or -1 on error with an accompanying 
   error message in statusMsg */

int
parseProtocolList(char *protocolStr, char **statusMsg)
{
    char *ps, *nextp;
    static char sMsg[300];
    int allowedProtos;
    
    allowedProtos = 0;
    
    sMsg[0] = '\0';
    *statusMsg = NULL;

    dbgMsg(DBG_CFG, "parseProtocolList: %s", protocolStr);

    ps = strdup(protocolStr);

    nextp = strtok(ps, " \t");

    while (nextp != NULL) {
	if (strcmp(nextp, PRNAME_WAP) == 0) {
	    allowedProtos |= PROTO_WAP;
	} else if(strcmp(nextp, PRNAME_HTTP) == 0) {
	    allowedProtos |= PROTO_HTTP;
	} else {
	    snprintf(sMsg, sizeof(sMsg), "Bad protocol name: %s", nextp);
	    *statusMsg = sMsg;
	} /* if */

        nextp = strtok(NULL, " \t");
    } /* while */

    free(ps);
    
    if (statusMsg == NULL && allowedProtos == 0)
	*statusMsg = "Missing protocol specification";
    
    return (*statusMsg == NULL) ? allowedProtos : -1;
} /* parseProtocolList */



/* Record location of an error in the parse file */

void
rcParseError(struct RcInfo *rcInfo, int lineNum, const char *msg)
{
    rcInfo->errorLine[rcInfo->numErrors] = lineNum;
    rcInfo->errorMsg[rcInfo->numErrors] = strdup(msg);
    
    rcInfo->numErrors++;
    
    dbgMsg(DBG_RCF, "%s parse error in line %d: %s", RCFILE, lineNum, msg);
} /* rcParseError */



/* getTokens() - a crude lexical analyser...

   Do an "in place" breakdown of the string pointed to by 'lp'
   to yield (in 'token') a set of up to 'maxToken' token strings.

   A token is either:

   	an unquoted sequence of non-whitespace characters, or

        a string delimited by single quotes.  The string may
        include nested whitespace, and use '\' as an escape
        character so that \' can be used to embed a single quote
        and \\ to embed a single \.
        The string is returned with delimiting quotes and
        backslash (\) escape characters removed.

        to allow the embedding of single quotes, and the backslash itself.

   NOTE: This routine modifies the string pointed to by 'lp'.

   Returns the number of tokens found, or a negative number (see GT_* 
   constants below) on error

   Example:  Parsing the line: sc greet 'echo \'Hello world\''
   would yield three tokens:
   	[sc]
        [greet]
        [echo 'Hello world']
*/

#define GT_UNCLOSED_STRING -1
#define GT_TOO_MANY_TOKENS -2

int
getTokens(char *lp, char **token, int maxTokens)
{
    int numTokens;
    char *pp;

    numTokens = 0;

    while (numTokens < maxTokens) {

    	/* Skip leading white space */

        while (isspace((int) *lp) && *lp != '\0')
            lp++;

        if (*lp == '\0') break;		/* No more tokens */

        if (*lp == '\'') { 		/* String token */
            lp++;
            token[numTokens] = lp;
            pp = lp;
            while (*lp != '\'' && *lp != '\0') {
            	if (*lp == '\\') {
                    if (*(lp+1) == 0) return GT_UNCLOSED_STRING;
                    lp++;
                } /* if */
                *pp++ = *lp++;
            } /* while */
            if (*lp == '\0') return GT_UNCLOSED_STRING;

            *pp = '\0';
            lp++;

        } else {			/* Normal token */
            token[numTokens] = lp;
            while (!isspace((int) *lp) && *lp != '\0')
                lp++;
            if (*lp != '\0') {
            	*lp = '\0';
                lp++;
            } /* if */
        } /* if */

        numTokens++;
    } /* for */

/*    dbgMsg(DBG_RCF, "getTokens found %d tokens", numTokens); */
    return (*lp == '\0') ? numTokens : GT_TOO_MANY_TOKENS;
} /* getTokens */



/* Take the set of init file parsing error messages saved in rcInfo, and 
   assemble into a single printable string which is suitable for display 
   to users.
*/

char *
buildRcErrorMsg(const struct RcInfo *rcInfo)
{
    int j;
    char buf[30];
    static char msgStr[5000];

    msgStr[0] = '\0';

    strcat(msgStr, "Error");
    if (rcInfo->numErrors > 1) 
        strcat(msgStr, "s");
    strcat(msgStr, " in initialisation file\n");

    for (j = 0; j < rcInfo->numErrors; j++) {
	if (rcInfo->errorLine[j] > 0) {
            sprintf(buf, "Line %d: ", rcInfo->errorLine[j]);
            strcat(msgStr, buf);
	} /* if */

        strcat(msgStr, rcInfo->errorMsg[j]);
        if (j != rcInfo->numErrors - 1)
            strcat(msgStr, "\n");
    } /* for */

    return msgStr;
} /* buildRcErrorMsg */



/* Process the init file 'set' command.  This has two forms:

   	set -o/+o option-name
	set NAME VALUE
	
   Parameters:
   	
   token and numToken: are the previously tokenised input line
   lineNum:	is the line number of this input line (used for inclusion
		in diagnostic messages
   rcInfo:	Is set to contain parse results (options and NAMES set by
		'set' command will be added to this structure)
   ifCondition: Determines whether we are currently inside
   		a conditional block whose condition is true or false.
		If the condition is false, we will parse this line (to check
		for errors) but not actually process it.
*/

void
processSetCmd(char *token[], int numTokens, int lineNum, 
        struct RcInfo *rcInfo, Boolean ifCondition)
{
    int opt;
    int j;

    if (numTokens != 3) {
        rcParseError(rcInfo, lineNum, 
		"Incorrect number of arguments for 'set'");
        return;
    } /* if */
    
    if (strchr("+-", token[1][0]) != NULL && token[1][1] == 'o') {
		
	/***** set -o/+o option-name */
	
        opt = 0;
        if (strcmp(token[2], OPTNAME_ALLOWSHELLCMD) == 0) {
            opt = OPT_ALLOWSHELLCMD;
        } else if (strcmp(token[2], OPTNAME_DISPLAYMENU) == 0) {
            opt = OPT_DISPLAYMENU;
        } else if (strcmp(token[2], OPTNAME_ALLOWSILENT) == 0) {
            opt = OPT_ALLOWSILENT;
        } else if (strcmp(token[2], OPTNAME_ALLOWCONTROLCHARS) == 0) {
            opt = OPT_ALLOWCONTROLCHARS;
        } else if (strcmp(token[2], OPTNAME_HISTORY) == 0) {
            opt = OPT_HISTORY;
        } else if (strcmp(token[2], OPTNAME_FILTERANSIESC) == 0) {
            opt = OPT_FILTERANSIESC;
        } else if (strcmp(token[2], OPTNAME_ALLOWUSERINIT) == 0) {
            opt = OPT_ALLOWUSERINIT;
        } else if (strcmp(token[2], OPTNAME_ALLOWTRIGRAPHS) == 0) {
            opt = OPT_ALLOWTRIGRAPHS;
        } else {
            rcParseError(rcInfo, lineNum, "Unknown option name for 'set'");
            return;
        } /* if */

	dbgMsg(DBG_RCF, "setOption: %s %d", token[2], opt);

        if (ifCondition) {
            if (token[1][0] == '-')
                rcInfo->setOptions |= opt;
            else
                rcInfo->setOptions &= ~opt;
        } /* if */
	
    } else {		/***** set NAME VALUE *****/
	
#define SN_ALLOWEDPROTOCOLS "allowedprotocols"
#define SN_OUTPUTBUFFERLIMIT "outputbufferlimit"
#define SN_OUTPUTWINDOWSIZE "outputwindowsize"
#define SN_HISTORYBLOCKSIZE "historyblocksize"
#define SN_SHORTCUTBLOCKSIZE "shortcutblocksize"
#define SN_SHELLTIMEOUT "shelltimeout"
#define SN_CSOUTPUTTIMEOUT "csoutputtimeout"
#define SN_CSMAXTRANSFERSIZE "csmaxtransfersize"
#define SN_WAPBROWSERSTYLE "wapbrowserstyle"

	char buf[100];

        char *setNames[] = { SN_ALLOWEDPROTOCOLS, SN_OUTPUTBUFFERLIMIT, 
	    		SN_CSMAXTRANSFERSIZE, SN_OUTPUTWINDOWSIZE,
	    		SN_HISTORYBLOCKSIZE, SN_SHORTCUTBLOCKSIZE,
			SN_SHELLTIMEOUT, SN_CSOUTPUTTIMEOUT, SN_WAPBROWSERSTYLE,
			NULL };
        char **sn;
	Boolean cmdOk;
	char *sName, *sValue;

	/* Search for NAME */
	
        for (sn = setNames; *sn != NULL; sn++) 
            if (strcmp(*sn, token[1]) == 0) 
                break;

        if (*sn == NULL) {
            rcParseError(rcInfo, lineNum, "Unknown variable name for 'set'");
            return;
	} /* if */
	
	sName = token[1];
	sValue = token[2];

	/* Check args: options other than 'wapbrowserstyle'  and 
	   'allowedprotocols' expect numeric arguments */

	if (strcmp(sName, SN_ALLOWEDPROTOCOLS) == 0) {
	    char *statusMsg;
	    int allowedProtos;
	    
	    allowedProtos = parseProtocolList(sValue, &statusMsg);
	    cmdOk = allowedProtos != -1;

	    /* Can't override config/cmd line specs */

	    allowedProtocols &= allowedProtos;

	} else if (strcmp(sName, SN_WAPBROWSERSTYLE) == 0) {
	    cmdOk =  strcmp(sValue, "up") == 0;

	} else if (strcmp(sName, SN_CSOUTPUTTIMEOUT) == 0) {
	    cmdOk = isFloatNumber(sValue);

	} else if (strcmp(sName, SN_OUTPUTBUFFERLIMIT) == 0) {
	    cmdOk = isIntNumber(sValue, 10);

	    /* Can't set limit higher than that specified at htshd startup */

	    if (cmdOk && atoi(sValue) > outputBufferLimit) {
		sprintf(buf, "%d", outputBufferLimit);
		sValue = buf;
	    } /* if */

	} else if (strcmp(sName, SN_OUTPUTWINDOWSIZE) == 0) {
	    cmdOk = isIntNumber(sValue, 10);
	    
	    if (cmdOk && atoi(sValue) < OUTPUT_WINDOW_SIZE_LOW_BOUND) {
		sprintf(buf, "%d", OUTPUT_WINDOW_SIZE_LOW_BOUND);
		sValue = buf;
	    } /* if */
	    
	    
	} else if (strcmp(sName, SN_HISTORYBLOCKSIZE) == 0) {
	    cmdOk = isIntNumber(sValue, 10);
	    if (cmdOk && atoi(sValue) < HISTORY_BLOCK_SIZE_LOW_BOUND) {
		sprintf(buf, "%d", HISTORY_BLOCK_SIZE_LOW_BOUND);
		sValue = buf;
	    } /* if */
	    
	} else if (strcmp(sName, SN_SHORTCUTBLOCKSIZE) == 0) {
	    cmdOk = isIntNumber(sValue, 10);
	    if (cmdOk && atoi(sValue) < SHORTCUT_BLOCK_SIZE_LOW_BOUND) {
		sprintf(buf, "%d", SHORTCUT_BLOCK_SIZE_LOW_BOUND);
		sValue = buf;
	    } /* if */

	} else {
	    cmdOk = isIntNumber(sValue, 10);
	} /* if */

	if (!cmdOk) {
            rcParseError(rcInfo, lineNum, "Invalid argument for 'set' command");
            return;
	} /* if */

	if (ifCondition) {

	    /* Add this name to the setName array in the RcInfo struct */
	    
            Boolean dupFnd;
	    int len;
	    int pos;
	    
	    /* Check to see whether this name is already in setName[].  If
	       it is, remove the previous definition and replace it with 
	       this one */
	    
	    dupFnd = FALSE;
	    
	    len = strlen(sName);
	    
	    for (j = 0; j < rcInfo->numSetVals && !dupFnd; j++) {
		if (strcmp(sName, rcInfo->setName[j]) == 0) {
		    dupFnd = TRUE;
		    pos = j;
		    dbgMsg(DBG_RCF, 
			    "Duplicate 'set Name' found: %s, at pos %d", 
			    sName, j);
		} /* if */
	    } /* for */

	    if (dupFnd) {	/* Remove previous definition */
		
		free(rcInfo->setName[pos]);
		free(rcInfo->setVal[pos]);
		
	    } else {		/* New definition, put at end of list */
		
		if (rcInfo->numSetVals >= MAX_RCFILE_SET_VALUES) {
                    rcParseError(rcInfo, lineNum, 
			    "Too many 'set NAME VALUE' commands");
                    return;
		} /* if */
		pos = rcInfo->numSetVals;    
                rcInfo->numSetVals++;    

	    } /* if */

	    /* Store definition in list */
	    
            rcInfo->setName[pos] = malloc(strlen(sName) + 1);
            rcInfo->setVal[pos] = malloc(strlen(sValue) + 1);
            if (rcInfo->setName[pos] == NULL || rcInfo->setVal[pos] == NULL) {
                rcParseError(rcInfo, lineNum, 
			"Memory allocation error processing 'set'");
                return;
            } /* if */
         
            strcpy(rcInfo->setName[pos], sName);
            strcpy(rcInfo->setVal[pos], sValue);

	    /* Certain options require processing on the server side... */
	    
	    if (strcmp(sName, SN_CSOUTPUTTIMEOUT) == 0) {
		csOutputTimeout = atof(sValue);
		if (csOutputTimeout < MIN_OUTPUT_TIMEOUT || 
			csOutputTimeout > MAX_OUTPUT_TIMEOUT) {
                    rcParseError(rcInfo, lineNum, 
			"csoutputtimeout argument is out of permitted range");
		    return;
		} /* if */
		
	    } else if (strcmp(sName, SN_CSMAXTRANSFERSIZE) == 0) {
		
		/* Ignore attempts to specify a value greater than that set
		   at startup time */

		csMaxTransferSize = min(atoi(sValue), csMaxTransferSize);

		/* And make sure it isn't set too low */

		csMaxTransferSize = max(csMaxTransferSize,
			CS_MAX_TRANSFER_SIZE_LOW_BOUND);	

	    } else if (strcmp(sName, SN_SHELLTIMEOUT) == 0) {
		
		shellTimeout = atoi(sValue);

	    } /* if */
        } /* if */
    } /* if */
} /* processSetCmd */



/* Handle the init file 'sc' command.
   
   Parameters:
   	
   token and numToken: are the previously tokenised input line
   lineNum:	is the line number of this input line (used for inclusion
		in diagnostic messages
   rcInfo:	Is set to contain parse results (the shortcut will be added
   		into this structure)
   ifCondition: Determines whether we are currently inside
   		a conditional block whose condition is true or false.
		If the condition is false, we will parse this line (to check
		for errors) but not actually process it.
*/
void
processScCmd(char *token[], int numTokens, int lineNum, 
        struct RcInfo *rcInfo, Boolean ifCondition)
{
    int nameToken;
    int defToken;
    int scOptions;

    if (numTokens < 2) {
        rcParseError(rcInfo, lineNum, "Wrong number of arguments for 'sc'");
        return;
    } /* if */

    scOptions = 0;

    if (token[1][0] == '-') {
        if (strchr(token[1], 'n') != NULL)
            scOptions |= SCOPT_NONL;
    } /* if */

/*                dbgMsg(DBG_RCF, "options=%o", scOptions); */

    nameToken = (token[1][0] == '-') ? 2 : 1;
    defToken = (nameToken + 1 < numTokens) ? (nameToken + 1) : nameToken;

    if (defToken >= numTokens) {
        rcParseError(rcInfo, lineNum, "Wrong Number of arguments for 'sc'");
        return;
    } /* if */

    if (ifCondition) {

	if (rcInfo->numSc >= MAX_RCFILE_SHORTCUTS) {
            rcParseError(rcInfo, lineNum, "Too many shortcuts");
            return;
	} /* if */

        rcInfo->scName[rcInfo->numSc] = strdup(token[nameToken]);
        rcInfo->scDef[rcInfo->numSc] = strdup(token[defToken]);
        rcInfo->scOptions[rcInfo->numSc] = scOptions;

        rcInfo->numSc++;
    } /* if */
} /* processScCmd */



/* Get a line of input from file named in 'fp'.  Line continuations using \
   prior to newline are handled by assembling multiple file lines (segments) 
   into a single input line.  (Except that lines considered to be comments 
   (i.e. first non-whitespace char is '#') cannot be continued in this way)

   Returns number of physical lines read from file, or 0 on EOF, or -1
   on error, in which case 'statusMsg' is left pointing to a string
   describing the error
*/

int
getLine(FILE *fp, int maxLength, char *line, int *lineLen, char **statusMsg)
{
    int numLines;		/* Total number of physical lines read 
				   to obtain this logical line */
    const int MAX_SEG_LEN = 10;
    char seg[MAX_SEG_LEN];	/* Next segment of input */
    int segLen;
    char *firstChar;		/* First non-whitepscae char in line (so 
				   we can tell if this is a comment line) */

    firstChar = line;
    numLines = 0;
    *lineLen = 0;

    while (numLines >= 0) {
	
	/* Read a line from file */

        if (fgets(seg, min(MAX_SEG_LEN, maxLength), fp) == NULL) {  /* EOF */

	    /* If we already have read some lines, then the previous line
	       must have had a \ continuation char before the newline - it's an
	       error to try and continue into EOF! */
	    
            if (numLines > 0) {
                *statusMsg = "Unexpected end of file";
		numLines = -1;
            } /* if */
            break;
	    
        } else {

	    segLen = strlen(seg);
	    if (seg[segLen-1] == '\n')
	        numLines++;

	    if (*lineLen + segLen + 1 > maxLength) {
                *statusMsg = "Line too long";
		numLines = -1;
        	*lineLen = 0;
	    } else {
		strcpy(&line[*lineLen], seg);
                *lineLen += segLen;
		
		/* Find first non-whitespace char in line */

		while (isspace((int) *firstChar))
		    firstChar++;

                if (line[*lineLen-1] == '\n') {	

		    /* Remove \newline (unless this is a comment line) */
		    
                    if (line[*lineLen-2] == '\\' && *firstChar != '#') {  
                        (*lineLen) -= 2;
			
                    } else { /* Otherwise, if hit a simple newline, then
				we're done! */
                        break;
                    } /* if */
                } /* if */
	    } /* if */
        } /* if */
    } /* while */

    line[*lineLen] = '\0';

    dbgMsg(DBG_RCF | DBG_CFG, "Getline returning %d  (lineLen=%d): %s", 
    		numLines, *lineLen, line); 
    return numLines;
} /* getLine */



/* processRcFile() - a cheap and cheerful initialisation file parser.

   Process the initialisation file (RCFILE) located in 'dir',
   returning parsed information in 'rcInfo'.  'statusMsg' returns
   a string describing the results of parsing - in the event of an error
   this message should be suitable for display to users, to give
   them some idea of the cause of the problem.

   Return TRUE if file was parsed okay, or if it simply didn't exist
   (the latter case can be distinguished by a NULL return for statusMsg).
*/

#define MAX_IF_LEVELS 20

void
processRcFile(const char *rcFile, struct RcInfo *rcInfo, 
	const char *clientProtocol, const char *userAgent, char **statusMsg)
{
    const int MAX_RCFILE_LINE_LEN = 2000;
    			/* Maximum length of lines in initialisation file */
    const int MAX_TOKENS_PER_RCFILE_LINE = 100;
    			/* Should be at least one greater than the largest
                           number of tokens we are allowed to have in a line
                           of an initialisation file */

    char line[MAX_RCFILE_LINE_LEN];
    			/* Current input line */
    int lineNum;	/* Current line number in initialisation file */
    int lineLen;	/* Total number characters in 'line' */
    char *token[MAX_TOKENS_PER_RCFILE_LINE];
  			/* Tokens in current input line */
    int numTokens;	/* Number of tokens in current line */
    FILE *fp;
    int ifLevel;	/* Level of 'if' nesting */
    Boolean ifCond[MAX_IF_LEVELS+1];
    			/* Truth/falsehood of 'if' at each nesting level */
    int j;
    char *glStatusMsg;
    int numLines;

    ifCond[0] = TRUE;
    ifLevel = 0;

    lineNum = 1;

    dbgMsg(DBG_RCF | DBG_CSH, "processRcFile: %s", rcFile);
    fp = fopen(rcFile, "r");
    if (fp == NULL) {
        *statusMsg = NULL;
        dbgMsg(DBG_RCF | DBG_CSH, 
		"Initialisation file not present: %s", rcFile);
    } else {

        /* Parse file a line at a time */

        for (numLines = 0, lineNum = 1; ; lineNum += numLines) {
            if (rcInfo->numErrors >= MAX_RCFILE_ERRORS) 
                break; 	/* rcInfo->numErrors is incremented by rcParseError */

	    numLines = getLine(fp, MAX_RCFILE_LINE_LEN, line, &lineLen, 
		    &glStatusMsg);
	    
	    if (numLines < 0) {
		rcParseError(rcInfo, lineNum + 1, glStatusMsg);
	        continue;
	    } /* if */

            if (lineLen == 0) break;		/* EOF */

            line[lineLen - 1] = '\0';	/* Remove trailing newline */

	    /* Skip comment lines */

	    for (j = 0; isspace((int) line[j]); j++);
	    if (line[j] == '#') continue;

	    /* Break line into tokens */

   	    numTokens = getTokens(line, token, MAX_TOKENS_PER_RCFILE_LINE);
            if (numTokens < 0) {
                rcParseError(rcInfo, lineNum, "Parse error");
                continue;
            } /* if */

            /* Skip empty lines */

            if (numTokens == 0) continue;

            /* Handle each type of command */

            if (strcmp(token[0], "set") == 0) {		/***** set *****/
                processSetCmd(token, numTokens, lineNum, rcInfo, 
			ifCond[ifLevel]);

            } else if (strcmp(token[0], "sc") == 0) {	/***** sc *****/
                processScCmd(token, numTokens, lineNum, rcInfo, 
			ifCond[ifLevel]);

            } else if (strcmp(token[0], "clearsc") == 0) { /***** clearsc *****/
		if (ifCond[ifLevel])
        	    rcInfo->numSc = 0;

            } else if (strcmp(token[0], "ifuseragent") == 0) { 
                                                /***** ifuseragent *****/
                if (numTokens < 2) {
                    rcParseError(rcInfo, lineNum,
                    		"Wrong number of arguments for 'ifuseragent'");
                    continue;
                } /* if */

                ifLevel++;
                ifCond[ifLevel] = FALSE;
                if (ifCond[ifLevel-1]) {
                    for (j = 1; j < numTokens && !ifCond[ifLevel]; j++)
                        ifCond[ifLevel] = fnmatch(token[j], userAgent, 0) == 0;
                } /* if */

            } else if (strcmp(token[0], "ifprotocol") == 0) { 
                                                /***** ifprotocol *****/
                if (numTokens != 2) {
                    rcParseError(rcInfo, lineNum,
                    		"Wrong number of arguments for 'ifprotocol'");
                    continue;
                } /* if */

                ifLevel++;
                if (ifCond[ifLevel-1]) 
                    ifCond[ifLevel] = 
				(strcmp(clientProtocol, PRNAME_WAP) == 0 && 
                    		    	strcmp(token[1], PRNAME_WAP) == 0) || 
                            	(strcmp(clientProtocol, PRNAME_HTTP) == 0 &&
                         	    	strcmp(token[1], PRNAME_HTTP) == 0);
                else
                    ifCond[ifLevel] = FALSE;

            } else if (strcmp(token[0], "fi") == 0) {	/***** fi *****/
                if (ifLevel <= 0) {
                    rcParseError(rcInfo, lineNum, "Mismatched 'fi'");
                    continue;
                } else {
                    ifLevel--;
                } /* if */

            } else {				/***** Unknown command *****/
                rcParseError(rcInfo, lineNum, "Unknown command");
                continue;
            } /* if */

        } /* for */

        fclose(fp);
    } /* if */

    dbgMsg(DBG_RCF, "processRcFile returning, numErrors=%d oprions=0x%x", 
	    rcInfo->numErrors, rcInfo->setOptions);

} /* processRcFile */



/* Process global and user specific initialisation files, if present */

Boolean
processInitFiles(const char *homeDir, struct RcInfo *rcInfo, 
	const char *clientProtocol, const char *userAgent, char **statusMsg)
{

    char rcPathname[PATH_MAX];
    			/* Pathname of initialisation file */
    
    initRcInfo(rcInfo);

    /* Certain options are on by default */

    rcInfo->setOptions = OPT_ALLOWSHELLCMD | OPT_DISPLAYMENU |
    		OPT_HISTORY | OPT_ALLOWCONTROLCHARS |
		OPT_ALLOWUSERINIT;

    if (logFileDir != NULL)
	rcInfo->setOptions |= OPT_SESSIONLOGGING;

    processRcFile(globalRcFile, rcInfo, clientProtocol, userAgent, statusMsg);

    if (rcInfo->numErrors > 0) {
	char buf[200];
	sprintf(buf, "Previous errors found in global initialisation file (%s)" 
		"; please contact your system administrator", GLOBAL_RC_FILE);
        rcParseError(rcInfo, 0, buf);
	
    } else if (rcInfo->setOptions & OPT_ALLOWUSERINIT) {	
	
	/* Only process user init file if no errors in global file,
	   and processing was not disabled in global init file */

        /* Build name of user initialisation file */

        sprintf(rcPathname, "%s/%s", homeDir, RCFILE);

        dbgMsg(DBG_RCF | DBG_CSH, "Processing user RC file: %s", rcPathname);

        processRcFile(rcPathname, rcInfo, clientProtocol, userAgent, statusMsg);
    } /* if */

    *statusMsg = (rcInfo->numErrors == 0) ? 
                "Initialisation file parsed okay" : buildRcErrorMsg(rcInfo);

    return rcInfo->numErrors == 0;
} /* processInitFiles */



        /***********************/
        /***** Child Shell *****/
        /***********************/


/* Send a message to HTTP server that login authentication / CCS /
   Child Shell startup failed */

void
sendLoginFailureMsg(int csd, const char *format, ...)
{
    va_list argList;
    char buf[1000];

    va_start (argList, format);

    dbgMsg(DBG_CCS | DBG_CSH, format, argList);

    vsnprintf(buf, sizeof(buf), format, argList);
    fdPrint(csd, "status=-1");
    fdPrint(csd, "msg=%s", buf);
    close(csd);
    va_end(argList);
} /* sendLoginFailureMsg */



/* Change real and effective UID and GID, as well as supplementary group list
   to be those for the user given by the pwd argument */

int
changeIdentity(const struct passwd *pwd)
{
    int ig, sg, su;

    dbgMsg(DBG_CSH, "Changing identity: %s", pwd->pw_name);

    ig = initgroups(pwd->pw_name, pwd->pw_gid);
    if (ig == -1) sysErrMsg("initgroups");
    sg = setgid(pwd->pw_gid);
    if (sg == -1) sysErrMsg("setgid");
    su = setuid(pwd->pw_uid);
    if (su == -1) sysErrMsg("setuid");

    return (ig == -1 || su == -1 || sg == -1) ? -1 : 0;
} /* changeIdentity */



/* Change environment variables, including those which should be
   specific to the user given in 'pwd' argument.
   Return 0 on success, or -1 on failure */

int
changeEnvironment(const char *homeDir, const char *shellPath, char *user,
        const char *clientProtocol, const char *userAgent)
{
    if (putenv("TERM=glasstty") == -1) return -1;	/* XXX Or "pty"? */
    if (setenv("HTSH_PROTOCOL", clientProtocol, 1) == 1) return -1;
    if (setenv("HTSH_USER_AGENT", userAgent, 1) == 1) return -1;
    if (setenv("HOME", homeDir, 1) == 1) return -1;
    if (setenv("SHELL", shellPath, 1) == 1) return -1;
    if (setenv("USER", user, 1) == 1) return -1;
    return 0;
} /* changeEnvironment */



/* Authenticate a login by reading a username+password(+client-type)
   from ccsInitSocket and checking them against the password/shadow
   password database, and the list of permitted users specified on the
   command line (if such a list was specified).

   Return parameters:
   pwd:	 	a pointer to the passwd record for this username
   authKey:	A randomly generated authorisation key (string of hex digits)
   		(space (MAX_AUTH_KEY_LEN+1 bytes) must be allocated for 
                this string by caller)
   clientProtocol: The protocol type (as specified by in authentication string)
   		(space (MAX_PROTOCOL_LEN+1 bytes) must be allocated for 
                this string by caller)
   userAgent:	The user agent
   		(space (MAX_USER_AGENT_LEN+1 bytes) must be allocated for this 
                string by caller)
   	

   Returns: one of the AU_* constants listed below
*/

#define AU_AUTH_SUCCEEDED   	0	/* Sucessfully authenticated */
#define AU_AUTH_FAILED 		-1	/* bad username/password */
#define AU_ACCESS_DENIED 	-2	/* user not on list of permitted logins
                                           (-u command line option to htshd)  */
#define AU_BAD_PROTOCOL_TYPE 	-3	/* Client type string was invalid
                                           (too long, or unrecognised) */
#define AU_BAD_USER_AGENT 	-5	/* User Agent string was invalid
                                           (too long) */
#define AU_ERROR 		-10	/* Error during authentication */


#define MAX_AUTH_KEY_LEN	50
#define MAX_PROTOCOL_LEN	50
#define MAX_USER_AGENT_LEN	250

int
authenticateUser(int ccsInitSocket, struct passwd **pwd, char *authKey,
        char *clientProtocol, char *userAgent)
{
    const int AUTH_BUFFER_SIZE = 1000;

    char buf[AUTH_BUFFER_SIZE+1];
    int numRead;
    char *password, *username, *cltype, *uagent, *encryptedPassword;
#ifdef USE_SHADOW_PWD
    struct spwd *spwd;
#endif
    Boolean loginOk;
    char *pwdp;
    int rt, ru;
#ifdef USE_DEV_RANDOM
    int fd;
#endif

    /* Get username and password */

    numRead = recv(ccsInitSocket, buf, AUTH_BUFFER_SIZE, 0);
    if (numRead == -1) {
    	sysErrMsg("recv");
        return AU_ERROR;
    } /* if */

    buf[numRead] = '\0';	/* Shouldn't be necessary, but be safe */
    dbgMsg(DBG_CSH, "read login char string %d chars", numRead);
    username = buf;
    password = buf + strlen(username) + 1;
    cltype = password + strlen(password) + 1;
    uagent = cltype + strlen(cltype) + 1;
    dbgMsg(DBG_CSH, "Username = %s; client type=%s; user agent=%s", 
            username, cltype, uagent);

    if (strcmp(cltype, PRNAME_WAP) == 0 || strcmp(cltype, PRNAME_HTTP)  == 0){
        strcpy(clientProtocol, cltype);
    } else {
        syslog(LOG_WARNING, "authenticateUser: Bad client protocol type: %s", 
		cltype);
        return AU_BAD_PROTOCOL_TYPE;
    } /* if */
    
    if (strlen(uagent) <= MAX_USER_AGENT_LEN) {
        strcpy(userAgent, uagent);
    } else {
        syslog(LOG_WARNING, "authenticateUser: Bad user agent type: %s", 
		uagent);
        return AU_BAD_USER_AGENT;
    } /* if */

    /* If we were supplied a list of permitted users, then check this
       user is in it */

    if (userSetSize > 0) {
        Boolean found;
        int j;

        dbgMsg(DBG_CSH, "Checking against list of permitted users");
        found = FALSE;
        for (j = 0; j < userSetSize && !found; j++)
            if (strcmp(userSet[j], username) == 0)
            	found = TRUE;
        if (!found) {
            syslog(LOG_INFO, 
		    "authenticateUser: User not permitted to login: %s", 
		    username);
            return AU_ACCESS_DENIED;
        } /* if */
    } /* if */

    /* Get user information from password and shadow password files */

    loginOk = FALSE;

    dbgMsg(DBG_CSH, "About to check password for %s", username);
    *pwd = getpwnam(username);
#ifdef USE_SHADOW_PWD
    dbgMsg(DBG_CSH, "About to check SHADOW password for %s", username);
    spwd = getspnam(username);
#endif

    if (*pwd == NULL
#ifdef USE_SHADOW_PWD
            || spwd == NULL 
#endif
    ) {
        dbgMsg(DBG_CSH, 
		"authenticateUser: Couldn't find username in password file");
        return AU_AUTH_FAILED;
    } /* if */

#ifdef USE_SHADOW_PWD
    pwdp = spwd->sp_pwdp;
#else
    pwdp = (*pwd)->pw_passwd;
#endif
    dbgMsg(DBG_CSH, "Home directory of %s is %s", username, (*pwd)->pw_dir);

    /* Check if supplied password is valid */

    encryptedPassword = crypt(password, pwdp);

    loginOk = strcmp(encryptedPassword, pwdp) == 0;
    dbgMsg(DBG_CSH, "Login %s", loginOk ? "succeeded" : "failed");
    if (!loginOk) {
        dbgMsg(DBG_CSH, "authenticateUser: Incorrect password");
        return AU_AUTH_FAILED;
    } /* if */

    dbgMsg(DBG_CSH, "Authentication succeeded");

    /* Generate authorisation key and write it in return parameter */

    /* freeBSD and Linux have /dev/urandom (not in Solaris though) */

#ifdef USE_DEV_RANDOM
#define RANDOM_DEV "/dev/urandom"

    dbgMsg(DBG_CSH, "Getting random values from %s", RANDOM_DEV);

    fd = open(RANDOM_DEV, O_RDONLY);
    if (fd == -1) {
	sysErrMsg("open - %s", RANDOM_DEV);
	return AU_ERROR;
    } /* if */
    if (read(fd, &rt, sizeof(rt)) == -1 || read(fd, &ru, sizeof(ru)) == -1) {
	sysErrMsg("read - %s", RANDOM_DEV);
	return AU_ERROR;
    } /* if */
   if (close(fd) == -1) sysErrMsg("close - %s", RANDOM_DEV);
#else
    srandom(time(NULL) % 100000000);
    rt = random();

    ru = ((((((pwdp[2] - 0x20) << 7) + (pwdp[3] - 0x20)) << 7) + 
                (pwdp[4] - 0x20)) << 7) + pwdp[5] - 0x20;

    srandom((rt ^ ru) % 100000000);
    ru = random();
#endif
    rt = abs(rt % 0x1000000);
    ru = abs(ru % 0x1000000);


    sprintf(authKey, "%06x%06x", rt, ru);

    dbgMsg(DBG_CSH, "Returning authentication key %s", authKey);

    /* If we got this far, everything went okay */

    return AU_AUTH_SUCCEEDED;
} /* authenticateUser */



/* Child Shell process main function

   After doing various initialisations, and opening the PTY slave used
   for communicating with CCS, send a message to the HTTP server to
   say we've successfully got going, and then exec a shell for this user.

   Input parameters:
   ccsInitSocket:	Descriptor of socket used for initialisation
   			of shell session.  After completing initialisation
                        this socket is used to write success/failure response
                        to the HTTP server.
   ptyName:		Name of the slave pty we are to use for
   			connection with CCS.  This routine opens the slave pty.
   ccsSocketPort:	Port number of the CCS listeing socket, which
   			we will notify to the HTPP server.
   pwd:			Password database record for the user running
   			this shell session
   authKey:		Previously generated authorization key, which we will
   			notify to HTTP server.
   clientProtocol	"wap" or "http"
   userAgent		User agent (browser type)
*/
void
childShell(int ccsInitSocket, const char *ptyName, int ccsSocketPort,
                const struct passwd *pwd, const char *authKey,
                char *clientProtocol, char *userAgent)
{
    int ptySlaveFd;	/* Descriptor for slave end of pty */
    char *shellPath;	/* Full path for user shell */
    char *shellName;	/* Filename (basename) of shell */
    int j;
    char *p;

    dbgMsg(DBG_CSH, "Child shell %d started for CCS %d", getpid(), getppid());

    /* Cease to be root: become the actual user */

    if (changeIdentity(pwd) == -1) {
        sendLoginFailureMsg(ccsInitSocket, "Couldn't change identity");
        exit(EXIT_FAILURE);
    } /* if */

    /* Get shell name from password record (default to Bourne shell if no shell
       specified), check that the shell really exists, and get its basename
       part */

    shellPath = pwd->pw_shell;
    if (shellPath == NULL) shellPath = "/bin/sh";
/* XXX This check will only work if shellPath is absolute (which it 
   should be in the ideal world.  However, programs such as login(1) work
   even when it is not, so for now we'll omit this check.  If the 
   shellPath doesn't point to an executable file, then the later exec 
   will fail in any case.

    if (access(shellPath, X_OK) == -1) {
        sendLoginFailureMsg(ccsInitSocket, 
		"Couldn't find shell: %s", shellPath);
        exit(EXIT_FAILURE);
    } 
*/

    p = strrchr(shellPath, '/');
    if (p == NULL) 
        p = shellPath;
    else 
        p++;
    
    shellName = malloc(strlen(p)+2);	
    				/* No need to free() cause we'll exec() later */
    strcpy(shellName+1, p);
    shellName[0] = '-';		/* Mark as login shell so that shell
    				   startup files get executed */

    dbgMsg(DBG_CSH, "Shell path=%s name=%s", shellPath, shellName);

    /* Change directory to location specified by password record */

    if (chdir(pwd->pw_dir) == -1) {
    	sysErrMsg("Child Shell: chdir: %s", pwd->pw_dir);
        sendLoginFailureMsg(ccsInitSocket, "Couldn't set directory to: %s",
                pwd->pw_dir);
        exit(EXIT_FAILURE);
    } /* if */

    /* Set environment variables */

    if (changeEnvironment(pwd->pw_dir, shellPath, pwd->pw_name, 
                clientProtocol, userAgent) == -1) {
        sendLoginFailureMsg(ccsInitSocket, "Couldn't change environment");
        exit(EXIT_FAILURE);
    } /* if */

    /* Put us in a separate session */
	
    if (setsid() == -1) {
    	sysErrMsg("childShell: setsid");
        sendLoginFailureMsg(ccsInitSocket, "setsid failed");
        exit(EXIT_FAILURE);
    } /* if */

    /* Open the slave end of the pty, make it our controlling tty,
       and dup all standard descriptors to it */

    ptySlaveFd = ptySlaveOpen(ptyName);
    if (ptySlaveFd == -1) {
        sendLoginFailureMsg(ccsInitSocket, "Can't open slave pseudo-terminal");
        exit(EXIT_FAILURE);
    } /* if */

#ifdef TIOCSTTY		/* Not needed for SysV derivates, such as Solaris */
    if (ioctl(ptySlaveFd, TIOCSCTTY, NULL) == -1) {
        sysErrMsg("childShell: ioctl TIOCSCTTY");
        sendLoginFailureMsg(ccsInitSocket,
                "Couldn't make pseudo-terminal into controlling terminal");
        exit(EXIT_FAILURE);
    } /* if */
#endif

    /* After this, we can't do any more output (e.g. dbgMsg()) as that 
       will appear on the pty and be collected by the CCS */

    if (dup2(ptySlaveFd, STDIN_FILENO) == -1 ||
    		dup2(ptySlaveFd, STDOUT_FILENO) == -1 ||
                dup2(ptySlaveFd, STDERR_FILENO) == -1) {
    	sysErrMsg("ChildShell: dup");
        sendLoginFailureMsg(ccsInitSocket,
                "Could not dup standard descriptors to pseudo-terminal");
        exit(EXIT_FAILURE);
    } /* if */

    if (ptySlaveFd > STDERR_FILENO)
        if (close(ptySlaveFd) == -1) sysErrMsg("childShell: close ptySlaveFd");

    /* If we got this far, then user was successfully authenticated, and
       we've got everything ready to go, so send the HTTP server all
       the info it needs to know, including the socket port for the CCS,
       and the authorization key */

    fdPrint(ccsInitSocket, "status=0");
    fdPrint(ccsInitSocket, "socketport=%d", ccsSocketPort);
    fdPrint(ccsInitSocket, "authkey=%s", authKey);
    fdPrint(ccsInitSocket, "shelltimeout=%d", shellTimeout);
    fdPrint(ccsInitSocket, "setoptions=%o", rcInfo.setOptions);
    fdPrint(ccsInitSocket, "numsc=%d", rcInfo.numSc);
    for (j = 0; j < rcInfo.numSc; j++)
        fdPrint(ccsInitSocket, "sc%d=%s", j, rcInfo.scName[j]);
    for (j = 0; j < rcInfo.numSetVals; j++)
        fdPrint(ccsInitSocket, "%s=%s", rcInfo.setName[j], rcInfo.setVal[j]);

    /* Now that we've sent the HTTP server the required information,
       we've finished with this socket, so get rid of it */

    if (close(ccsInitSocket) == -1)
        sysErrMsg("childShell: close ccsInitSocket");

    /* And now we're ready to kick off the shell */

    execlp(shellPath, shellName, NULL);
    sysErrExit("execlp: %s", shellPath);
} /* childShell */



        /********************************/
        /***** Child Command Server *****/
        /********************************/


/* Sometimes, the CCS just has to die.  In the process we also kill the child
   shell and wait a few seconds for it to be reaped by the SIGCHLD catcher
   before we exit.  */

void
killChildShellAndDie(pid_t childShellPid)
{
    int cnt;
    
    kill(childShellPid, SIGKILL);
    syslog(LOG_ERR, "Child command server and shell exiting "
	    "(killChildShellAndDie)");

    cnt = 0;	/* Child shouldn't take long to die, but just in case...  */
    while (childShellAlive && cnt < 10) {
	dbgMsg(DBG_CCS | DBG_CSH, "CCS %d: waiting for child shell %d to die",
		getpid(), childShellPid);
	sleep(1);
	cnt++;
    } /* while */

    exit(EXIT_FAILURE);
} /* killChildShellAndDie */



/* If we encounter an error reading from the CCS Command Socket,
   we give up and go home */

void
ccsCmdSocketReadFailExit(int numRead, pid_t childShellPid, const char *msg)
{
    if (numRead == -1)
    	sysErrMsg("CCS: read on command socket: %s", msg);
    else if (numRead == 0)
    	syslog(LOG_ERR, "CCS: unexpected EOF on command socket: %s", msg);
    else
    	syslog(LOG_ERR, "CCS: partial read (%d bytes) on command socket: %s",
                numRead, msg);
    killChildShellAndDie(childShellPid);
} /* ccsCmdSocketReadFailExit */



/* If we encounter an error writing to the pty master,
   we give up and go home */

void
ptyMasterWriteFailExit(int numWritten, pid_t childShellPid, const char *msg)
{
    if (numWritten == -1)
    	sysErrMsg("CCS: write on master pty: %s", msg);
    else
    	syslog(LOG_ERR, "CCS: partial write (%d bytes) on master pty: %s",
                numWritten, msg);
    killChildShellAndDie(childShellPid);
} /* ptyMasterWriteFailExit */



/* Return true if the IP address provided in 'host' is one of the 
   HTTP Servers recorded in 'httpServers[]' and thus permitted to start 
   shells on this system */

Boolean
isPermittedHttpServer(long host)
{
    Boolean isOk;
    int j;

    /* If no HTTP Server Host list was specified, then we only
       accept logins from localhost */
    
    if (numHttpServers == 0) {
        isOk = host == inet_addr("127.0.0.1");
    } else {
        isOk = FALSE;
        for (j = 0; j < numHttpServers && !isOk; j++) 
            isOk = httpServer[j] == host;
    } /* if */
    
    dbgMsg(DBG_CSH | DBG_CCS, "isPermittedHttpServer returning: %d", isOk);
    return isOk;
} /* isPermittedHttpServer */



/* Accept a new connection on ccsSocket, verify that the 
   connection is from a permitted server, and read and check 
   the authorisation key (which must be the first piece of 
   data arriving on the connected socket).

   Returns the descriptor of the new connected socket, or one of the 
   negative ACC_* values listed below if something goes wrong (in which
   case this routine closes the connected socket)

   If we timeout waiting for a connection, or if things go badly wrong, 
   we kill child shell, and exit.
*/
#define ACC_BAD_AUTH_KEY   -2	/* Supplied authorisation key is incorrect */
#define ACC_UNEXPECTED_EOF -3	/* Hit EOF while trying to read 
                                   authorisation key */
#define ACC_ACCESS_DENIED  -4	/* Not accepting connections from this server */

int acceptCcsConnection(int ccsSocket, pid_t childShellPid, char *authKey)
{
    struct sockaddr_in clientAddr;	/* Address of connecting client */
    int addrLen;
    fd_set readFds;
    struct timeval timeout;
    int s;
    int numRead;
    int ccsCmdSocket;
    int authKeyLen;		/* Length of authKey */
    char cmdAuthKey[MAX_AUTH_KEY_LEN + 1]; 
    				/* Authorisation key supplied by HTTP server */

    authKeyLen = strlen(authKey);

    /* Wait for a connection */

    dbgMsg(DBG_CCS, "CCS %d waiting for connection on ccsSocket %d", 
	    getpid(), ccsSocket);

    FD_ZERO(&readFds);
    FD_SET(ccsSocket, &readFds);
    timeout.tv_sec = shellTimeout;
    timeout.tv_usec = 0;

    s = select(ccsSocket + 1, &readFds, NULL, NULL, &timeout);
    
    if (s == -1) { /* May fail because of SIGCHLD handler (i.e. CSh has died) */
        sysErrMsg("select");
        killChildShellAndDie(childShellPid);
    } /* if */
    
    if (s == 0) {
        syslog(LOG_INFO, "CCS %d timed out on connection - exiting", getpid());
        killChildShellAndDie(childShellPid);
    } /* if */
    
    /* Accept the conection */

    addrLen = sizeof(clientAddr);
    ccsCmdSocket = accept(ccsSocket, (struct sockaddr *) &clientAddr, &addrLen);
    if (ccsCmdSocket == -1) {
        /* If the accept() fails, there's little we can do to
           inform the client */

        sysErrMsg("accept");
        killChildShellAndDie(childShellPid);
    } /* if */
    
    dbgMsg(DBG_CCS, "CCS %d accepted conection from %s", getpid(),
           inetAddressString(&clientAddr));

    /* Check that we are permitting connections from this HTTP server */

    if (!isPermittedHttpServer(clientAddr.sin_addr.s_addr)) {
        char *msg = "Connections not permitted from this HTTP server";
        syslog(LOG_INFO, "acceptCcsConnection: Denied access from %s",
                inetAddressString(&clientAddr));
        writen(ccsCmdSocket, &RESP_FATAL, 1);
        writen(ccsCmdSocket, msg, strlen(msg));
        close(ccsCmdSocket);
        return ACC_ACCESS_DENIED;
    } /* if */

    /* Read and check authorisation key */
    
    numRead = readn(ccsCmdSocket, &cmdAuthKey, authKeyLen);
    if (numRead == 0) {
        close(ccsCmdSocket);
        return ACC_UNEXPECTED_EOF;
    } /* if */
    
    if (numRead != authKeyLen)
        ccsCmdSocketReadFailExit(numRead, childShellPid, "cmdAuthKey");

    cmdAuthKey[authKeyLen] = '\0';
    dbgMsg(DBG_CCS, "Authorisation key = %s", cmdAuthKey);

    if (strncmp(cmdAuthKey, authKey, authKeyLen) != 0) {
        char *msg = "Bad authorisation key";

        syslog(LOG_INFO, "acceptCcsConnection: Bad authkey: %s (expected %s)",
                cmdAuthKey, authKey);
        writen(ccsCmdSocket, &RESP_ERROR, 1);
        writen(ccsCmdSocket, msg, strlen(msg));
        close(ccsCmdSocket);
        return ACC_BAD_AUTH_KEY;
    } /* if */
    
    /* If we got this far, we succeeded, so return connected socket */
    
    return ccsCmdSocket;
} /* acceptCcsConnection */



/* Read a command from the connected socket, and pass the appropriate 
   input to the child shell
*/
Boolean
processShellInputCmd(int ccsCmdSocket, pid_t childShellPid, int ptyMasterFd)
{
    const int CMD_NUM_ARG_LEN = 4; /* Number of characters in numeric argument
                                   which forms part of some CMD_* messages */
    char cmdNumArgStr[CMD_NUM_ARG_LEN + 1];
    				/* Numeric argument immediately following
       				   command char (only used for some commands) */
    int cmdNumArg;		/* Integer form of cmdNumArgStr */
    char cmdBuf[1000];		/* Should be big enough for largest command
                                   from HTTP server */
    char controlChar[1];	/* For CMD_CONTROL and CMD_SIGNAL:
                                   specifies control character to be
                                   written to PTY.  For CMD_CONTROL, this is
                                   specified in printable form.  Thus
                                   Control-C (0x03) is specified using
                                   the letter 'C' (0x43).  For CMD_SIGNAL,
                                   the signal is specified via cdNumArgStr */
    char cmdChar;		/* Determines type of this command */
    int numRead, numWritten;

    /* Read the command char, to find out what kind of command we have */

    numRead = readn(ccsCmdSocket, &cmdChar, 1);
    if (numRead == 0)
        goto unexpectedEof;
    if (numRead != 1)
        ccsCmdSocketReadFailExit(numRead, childShellPid, "cmdChar");

    dbgMsg(DBG_CCS, "Got cmdChar: %c", cmdChar);

    /* Read further input, depending on the type of command char */

    if (cmdChar == CMD_CONTROL) {
        numRead = readn(ccsCmdSocket, &controlChar, 1);
        if (numRead == 0)
            goto unexpectedEof;
        if (numRead != 1)
            ccsCmdSocketReadFailExit(numRead, childShellPid, "controlChar");
        controlChar[0] = toupper(controlChar[0]);

        dbgMsg(DBG_CCS, "######Control char: %c %d", controlChar[0],
               controlChar[0] - 0x40);

        controlChar[0] = controlChar[0] - 0x40;

    } else if (cmdChar == CMD_SIGNAL || cmdChar == CMD_SHELLCMD ||
                 cmdChar == CMD_SHORTCUT) {

	numRead = readn(ccsCmdSocket, &cmdNumArgStr, CMD_NUM_ARG_LEN);
        if (numRead == 0) goto unexpectedEof;
        if (numRead != CMD_NUM_ARG_LEN)
            ccsCmdSocketReadFailExit(numRead, childShellPid, "cmdNumArgStr");
        cmdNumArgStr[CMD_NUM_ARG_LEN] = '\0';
        cmdNumArg = atoi(cmdNumArgStr);
	if (cmdNumArg < 0) {
            syslog(LOG_WARNING, 
		    "CCS:  processShellInputCmd: negative cmdNumArg %d", 
		    cmdNumArg);
            return FALSE;
	} /* if */

        dbgMsg(DBG_CCS, "cmdNumArg = %d", cmdNumArg);

        if (cmdChar == CMD_SHELLCMD) {
            if (cmdNumArg == 0) {
                dbgMsg(DBG_CCS, "Zero length command");
                numRead = 0;
            } else {
                numRead = readn(ccsCmdSocket, &cmdBuf, cmdNumArg);
                if (numRead == 0) goto unexpectedEof;
                if (numRead != cmdNumArg)
                    ccsCmdSocketReadFailExit(numRead, childShellPid,
                                            "cmdBuf");
                cmdBuf[numRead] = '\0';
                dbgMsg(DBG_CCS, "CCS cmdBuf = %s", cmdBuf);
            } /* if */
        } /* if */
    } /* if */
    
    /* Interpret the command from the HTTP server and send appropriate
       data to Child Shell via pty */
    
    switch (cmdChar) {
    
    case CMD_CLOSE:
        close(ptyMasterFd);
        dbgMsg(DBG_CCS, "CCS %d closing", getpid());
        killChildShellAndDie(childShellPid);
        exit(EXIT_SUCCESS);
        break;

    case CMD_CONTROL:
        numWritten = writen(ptyMasterFd, controlChar, 1);
        if (numWritten != 1)
            ptyMasterWriteFailExit(numWritten, childShellPid, "controlChar");
        break;

    case CMD_SIGNAL:
        controlChar[0] = (cmdNumArg == SIGQUIT) ? '' :
            (cmdNumArg == SIGSTOP) ? '' : '';
        numWritten = writen(ptyMasterFd, controlChar, 1);
        if (numWritten != 1)
            ptyMasterWriteFailExit(numWritten, childShellPid,
                                  "controlChar - CMD_SIGNAL");
        break;

    case CMD_NULL:
        /* Null command - we just want to get some more output */
        break;

    case CMD_SHELLCMD:
        if (cmdNumArg > 0) {
            cmdBuf[cmdNumArg] = '\0';
            dbgMsg(DBG_CCS, "CCS: Sending shell command to pty: %s", cmdBuf);
            numWritten = writen(ptyMasterFd, cmdBuf, cmdNumArg);
            if (numWritten != cmdNumArg)
                ptyMasterWriteFailExit(numWritten, childShellPid, "cmdBuf");
        } /* if */
        break;

    case CMD_SHORTCUT:

	/* Shortcut: we get a shortcut number, and we must send 
	   corresponding shortcut text to shell */
	
        if (cmdNumArg >= rcInfo.numSc) {
            dbgMsg(DBG_CCS, "Shortcut command out of range: %d (max=%d)",
                   cmdNumArg, rcInfo.numSc);
        } else {
            int len;

            dbgMsg(DBG_CCS, "Got shortcut command: %d", cmdNumArg);

            len = strlen(rcInfo.scDef[cmdNumArg]);
            numWritten = writen(ptyMasterFd, rcInfo.scDef[cmdNumArg], len);
            if (numWritten != len)
                ptyMasterWriteFailExit(numWritten, childShellPid, "shortcut");

            if ((rcInfo.scOptions[cmdNumArg] & SCOPT_NONL) == 0) {
                numWritten = writen(ptyMasterFd, "\n", 1);
                if (numWritten != 1)
                    ptyMasterWriteFailExit(numWritten, childShellPid, 
			    "newline");
            } /*if */
        } /* if */
        break;

    default:
        syslog(LOG_WARNING, 
		"CCS:  processShellInputCmd: BAD CASE - cmdChar: %c", cmdChar);
        return FALSE;

    } /* switch */

    return TRUE;
    
unexpectedEof:
    syslog(LOG_NOTICE, "CCS: processShellInputCmd: "
	    "Got EOF part way through command: discarding");

    return FALSE;
} /* processShellInputCmd */



/* Read output from the child shell (on master side of pty) and transmit 
   to HTTP server (via ccsCmdSocket) */

void
sendShellOutputToHttpServer(int ccsCmdSocket, pid_t childShellPid,
                            int ptyMasterFd, FILE *logFile)
{
    int numRead;
    int totWritten;		/* Total bytes written to HTTP server */
    struct timeval timeout;
    fd_set readFds;
    time_t startTime;		/* Time when we began reading this burst 
				   of output */
    time_t endTime;		/* Upper limit time on processing output */

    dbgMsg(DBG_CCS | DBG_OUT, "About to check output from pty");

    startTime = time(NULL);

    if (write(ccsCmdSocket, &RESP_NORMAL, 1) == -1) {
        errorMsg("CCS: write - RESP_NORMAL - failed");
        killChildShellAndDie(childShellPid);
    } /* if */

    /* Loop reading blocks of output from shell, end sending to HTTP Server.
       Stop if we have read 'csMaxTransferSize' bytes, or we have spent
       a large total time reading output (3 * 'csOutputTimeout').
       In addition, this loop may exit if any single read timesout
       after 'csOutputTimeoutSecs' */

    endTime = max(startTime + (time_t) (csOutputTimeout * 3), startTime + 1);
    	/* Upper limit is always at least one second greater than start time */

    for (totWritten = 0; totWritten < csMaxTransferSize && 
	    time(NULL) < endTime;) {

        FD_ZERO(&readFds);
        FD_SET(ptyMasterFd, &readFds);

        /* We wait a limited time for some output from the Child Shell.
           If no input is received in that time, then we cease reading
           and close the ccsCmdSocket.  If any further output does arrive
           it will be returned when the next command is processed.  */

        timeout.tv_sec = (int) csOutputTimeout;
        timeout.tv_usec = (int) ((csOutputTimeout) * 1000000) % 1000000;

        if (select(max(ptyMasterFd, ccsCmdSocket) + 1, &readFds,
                   NULL, NULL, &timeout) == -1) {
            sysErrMsg("CCS: select");
            killChildShellAndDie(childShellPid);
        } /* if */
	
        if (FD_ISSET(ptyMasterFd, &readFds)) {
            char buf[100];
	    int numToRead;

	    numToRead = min(sizeof(buf)-1, csMaxTransferSize - totWritten);

            numRead = read(ptyMasterFd, buf, numToRead);
            if (numRead == -1) {
                /* XXX should we handle this better?  For example
		   send a special message type to HTTP Server? */
                sysErrMsg("CCS: read - ptyMasterFd");
                killChildShellAndDie(childShellPid);
            } /* if */

            if (numRead == 0)
                break;

            buf[numRead] = '\0';	/* Just done for debugging msg */
            dbgMsg(DBG_OUT, "Pty->socket [%2dB] %s", numRead, buf);

	    if (logFile != NULL) {
	        fprintf(logFile, "%s", buf);
	        fflush(logFile);
	    } /* if */

            if (write(ccsCmdSocket, buf, numRead) == -1) {
                sysErrMsg("CCS: write - ccsCmdSocket");
                killChildShellAndDie(childShellPid);
            } /* if */
            totWritten += numRead;
	    
        } else {	/* Timeout waiting for further output: we're done */
            
    	    dbgMsg(DBG_CCS | DBG_OUT, 
		    "Timed out (%5.3f) while waiting for shell output", 
		    csOutputTimeout);
            break;
	    
        } /* if */

    } /* for */
    dbgMsg(DBG_CCS | DBG_OUT, "Finished output processing");

} /* sendShellOutputToHttpServer */



/* The Child Command Server (CCS)

   Just loop continuously, getting a connection from the HTTP server,
   reading a single command, interpreting and sending it via the pty
   to the child shell, reading the (pty) output from the child, and
   sending this output back to the HTTP server.

   Parameters:

   ccsSocket:	Descriptor of the listening socket on which we accept incoming
   		connections from HTTP server
   ptyMasterFd:	Descriptoy of the pty which connects us to child shell
   childShellPid: PID of the child shell
   authKey:	Previously generated athorisation key which must appear
   		at start of each command we receive from HTTP server
*/

void
childCommandServer(int ccsSocket, int ptyMasterFd, pid_t childShellPid,
                   char *authKey, struct passwd *pwd, char *clientProtocol)
{
    int ccsCmdSocket;		/* Descriptor of current accepted connection */
    FILE *logFile = NULL;
    char logFileName[PATH_MAX];
    struct utsname utsb;
    struct tm *tmp;
    time_t tp;

    dbgMsg(DBG_CCS, "CCS entered childCommandServer()");

    if (logFileDir != NULL) {
        if (uname(&utsb) == -1) {
	    sysErrMsg("utsname");
        } else {
	    tp = time(NULL);
	    tmp = localtime(&tp);
            sprintf(logFileName, "%s/%s_%s_%s_%02d%02d%02d.%02d%02d%02d.log", 
		    logFileDir, 
		    utsb.nodename, pwd->pw_name, 
		    clientProtocol,
		    tmp->tm_year % 100, tmp->tm_mon+1, tmp->tm_mday,
		    tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
    
	    logFile = fopen(logFileName, "a");
	    chmod(logFileName, 0600);

	    dbgMsg(DBG_CCS | DBG_CSH, "script logFile: %s", logFileName);
        } /* if */
    } /* if */

    while (childShellAlive) {	/* Global may be changed by SIGCHLD handler */

        /* Get a new connection */

        ccsCmdSocket = acceptCcsConnection(ccsSocket, childShellPid, authKey);
        if (ccsCmdSocket < 0)
            continue;

        /* Read command from HTTP server and send appropriate input
	   to child shell */

        if (processShellInputCmd(ccsCmdSocket, childShellPid, 
		    	ptyMasterFd)) {

            /* Read output from pseudoterminal, and send back to HTTP server
               using CCS Command Socket */

            sendShellOutputToHttpServer(ccsCmdSocket, childShellPid,
                                        ptyMasterFd, logFile);
        } /* if */

	dbgMsg(DBG_CCS, "CCS: About to close ccsCmdSocket: %d", ccsCmdSocket);
        if (close(ccsCmdSocket) == -1) {
            sysErrMsg("CCS close ccsCmdSocket");
            killChildShellAndDie(childShellPid);
        } /* if */
    } /* while */

    dbgMsg(DBG_CCS, "CCS %d all done ", getpid());
    exit(EXIT_SUCCESS);
} /* childCommandServer */



pid_t globChildShellPid;

void
clientShellTerm(int sigNum)
{
    dbgMsg(DBG_CCS | DBG_CSH, "CCS %d: Caught signal %d, exiting", 
	    getpid(), sigNum);
    killChildShellAndDie(globChildShellPid);
} /* clientShellTerm */



void
sendAuthFailureMsg(int authStatus, int ccsInitSocket)
{
    
    dbgMsg(DBG_CCS | DBG_CSH, "Authentication failed: %d", authStatus);

    switch (authStatus) {

    case AU_AUTH_FAILED:
        sendLoginFailureMsg(ccsInitSocket, 
		    "Username/password incorrect");
	break;

    case AU_ACCESS_DENIED:
        sendLoginFailureMsg(ccsInitSocket, 
		    "User not allowed to login via htsh");
	break;

    case AU_BAD_PROTOCOL_TYPE:
        sendLoginFailureMsg(ccsInitSocket, "Bad client protocol type");
	break;

    case AU_BAD_USER_AGENT:
        sendLoginFailureMsg(ccsInitSocket, "Bad user agent");
	break;

    case AU_ERROR:
        sendLoginFailureMsg(ccsInitSocket, "Authentication failure");
	break;

    default:
        sendLoginFailureMsg(ccsInitSocket, 
		    "Unknown user authentication error %d", authStatus);
	break;

    } /* switch */
} /* sendAuthFailureMsg */


/* startShellSession

   Start a new shell session. This requires a Child Command Server (CCS)
   process (which will be this process), and a Child Shell (CSh) process
   which we will create in this routine.

   We authenticate the user, read their initialisation file (~/RCFILE),
   open the PTY master which will be used for connection between CCS and CSh,
   and bind a socket which will be used by CCS to accept future connections
   from the HTTP server.  Then we fork a child to be the CSh, while
   parent passes control to the childCommandServer() routine.

   Input parameters:
   ccsInitSocket:	The socket connection to the HTTP server
   			created by accept() in the MCS.  This will supply
                        us a username/password, and be used to write a
                        success/failure message back to the HTTP server.
*/
void
startShellSession(int ccsInitSocket)
{
    pid_t childShellPid;
    int ptyMasterFd;
    int ccsSocketPort;
    int ccsSocket;
    struct sockaddr_in ccsSocketAddr;
    int addrLen;
    int authStatus;
    struct passwd *pwd;
    char *statusMsg;
    char authKey[20];
    char clientProtocol[MAX_PROTOCOL_LEN+1];
    				/* wap or http */
    char userAgent[MAX_USER_AGENT_LEN+1];
    				/* Browser type as returned by 
				   HTTP_USER_AGENT */
    sigset_t blockSet, oldSet;
    struct sockaddr_in clientAddr;
    struct hostent *hostInfo;
    char *hostname;


    /* Authenticate the user */

    authStatus = authenticateUser(ccsInitSocket, &pwd, authKey, 
            			clientProtocol, userAgent);

    if (authStatus < 0) {
	sendAuthFailureMsg(authStatus, ccsInitSocket);
        exit(EXIT_FAILURE);
    } /* if */

    /* Find out name of remote host */

    addrLen = sizeof(clientAddr);
    if (getpeername(ccsInitSocket, (struct sockaddr *) &clientAddr, 
		&addrLen) == -1) {
        sendLoginFailureMsg(ccsInitSocket, "Can't get peer socket address");
        exit(EXIT_FAILURE);
    } /* if */

    hostInfo = gethostbyaddr((char *) &clientAddr.sin_addr, 
	    sizeof(clientAddr.sin_addr), clientAddr.sin_family);
    if (hostInfo == NULL)
	hostname = inet_ntoa(clientAddr.sin_addr);
    else
	hostname = hostInfo->h_name;  /* point to host's name */
	
    strncpy(remoteHost, hostname, UT_HOSTSIZE);

    dbgMsg(DBG_CCS, "Remote host is: /%s/ %d", remoteHost, UT_HOSTSIZE);
    
    /* Process the initialisation files */

    if (!processInitFiles(pwd->pw_dir, &rcInfo, clientProtocol, userAgent, 
		&statusMsg)) {
        syslog(LOG_INFO, "Error processing rcFile: %s", statusMsg);
        sendLoginFailureMsg(ccsInitSocket, statusMsg);
        exit(EXIT_FAILURE);
    } /* if */

    /* Are we allowed to login using this protocol */

    dbgMsg(DBG_CCS, "Checking protocol: %s %d", clientProtocol, 
	    allowedProtocols);

    if ((strcmp(clientProtocol, PRNAME_WAP) == 0 && 
	    	(allowedProtocols & PROTO_WAP) == 0) || 
    	(strcmp(clientProtocol, PRNAME_HTTP) == 0 && 
	    	(allowedProtocols & PROTO_HTTP) == 0)) {
        sendLoginFailureMsg(ccsInitSocket, 
		"Can't login using this protocol (%s)", clientProtocol);
        exit(EXIT_FAILURE);
    } /* if */

    /* printRcInfo(&rcInfo); */

    /* Open the master side of pty to be used to communicate with Child Shell */

    ptyMasterFd = ptyMasterOpen(ptyName);
    if (ptyMasterFd == -1) {
        sendLoginFailureMsg(ccsInitSocket, "Can't open master pseudo-terminal");
        exit(EXIT_FAILURE);
    } /* if */

    dbgMsg(DBG_CCS | DBG_CSH, "Slave is: %s", ptyName);

    /* Before we change our identity set ownership and permissions on
       pty slave */

    if (setPtyOwnerAndMode(ptyName, pwd) == -1) {
        sendLoginFailureMsg(ccsInitSocket, 
		"Couldn't set pseudo-tty ownership and permissions");
        exit(EXIT_FAILURE);
    } /* if */

    /* Create a socket which will be used by the HTTP server to
       to connect to the CCS */

    ccsSocket = socket(AF_INET, SOCK_STREAM, 0);
    if (ccsSocket == -1) {
        sysErrMsg("startShellSession: socket: ccsSocket");
        sendLoginFailureMsg(ccsInitSocket, 
		"Couldn't create child server socket");
        exit(EXIT_FAILURE);
    } /* if */

    /* By binding the socket, we can then find out which
       port number it is available on - we need to communicate this
       to the HTTP server session so that it knows where to find the CCS */

    ccsSocketAddr.sin_family = AF_INET;
    ccsSocketAddr.sin_addr.s_addr = INADDR_ANY;
    ccsSocketAddr.sin_port = 0;
    if (bind(ccsSocket, (struct sockaddr *) &ccsSocketAddr, 
		sizeof(struct sockaddr_in)) == -1) {
        sysErrMsg("startShellSession: bind: ccsSocket");
        sendLoginFailureMsg(ccsInitSocket, "Couldn't bind child server socket");
        exit(EXIT_FAILURE);
    } /* if */

    dbgMsg(DBG_CCS | DBG_CSH, "CCS socket after bind=%s", 
	    socketAddressString(ccsSocket));

    addrLen = sizeof(ccsSocketAddr);
    if (getsockname(ccsSocket, (struct sockaddr *) &ccsSocketAddr, 
		&addrLen) == -1) {
        sysErrMsg("startShellSession: getsockname: ccsSocket");
        sendLoginFailureMsg(ccsInitSocket, "Couldn't get name of child socket");
        exit(EXIT_FAILURE);
    } /* if */

    ccsSocketPort = ntohs(ccsSocketAddr.sin_port);
    dbgMsg(DBG_CCS | DBG_CSH, "CCS Socket port is %d", ccsSocketPort);

    if (signal(SIGCHLD, childShellReaper) == SIG_ERR) {
    	sysErrMsg("CCS: signal - SIGCHLD");
        sendLoginFailureMsg(ccsInitSocket,
                "Child Command Server couldn't set SIGCHLD handler");
        exit(EXIT_FAILURE);
    } /* if */

    /* After a SIGTERM or SIGINT, we exit gracefully */

    if (signal(SIGTERM, clientShellTerm) == SIG_ERR) {
    	sysErrMsg("CCS: signal - SIGTERM");
        sendLoginFailureMsg(ccsInitSocket,
                "Child Command Server couldn't set SIGTERM handler");
        exit(EXIT_FAILURE);
    } /* if */

    if (signal(SIGINT, clientShellTerm) == SIG_ERR) {
    	sysErrMsg("CCS: signal - SIGINT");
        sendLoginFailureMsg(ccsInitSocket,
                "Child Command Server couldn't set SIGINT handler");
        exit(EXIT_FAILURE);
    } /* if */

    /* It looks like persistent sockets (pfsockopen()) don't work in PHP,
       hence we've got to create a listening socket and open a new
       connection every time... */

    if (listen(ccsSocket, 1) == -1) {
    	sysErrMsg("CCS: listen");
        sendLoginFailureMsg(ccsInitSocket,
                "Child Command Server couldn't establish listening socket");
        exit(EXIT_FAILURE);
    } /* if */

    /* Create process for Child Shell */

    /* We block SIGCHLD, until parent has created utmp and wtmp entries */
    
    sigemptyset(&blockSet);
    sigaddset(&blockSet, SIGCHLD);

    if (sigprocmask(SIG_BLOCK, &blockSet, &oldSet) == -1) {
	sysErrMsg("CCS: sigprocmask SIGBLOCK");
        sendLoginFailureMsg(ccsInitSocket,
                "Child Command Server couldn't block SIGCHLD");
        exit(EXIT_FAILURE);
    } /* if */

    childShellAlive = TRUE;	/* Must set here, rather than after fork()
    				   to avoid race condition */
    childShellPid = fork();
    

    switch (childShellPid) {

    case 0:		/* CHILD */
	dbgMsg(DBG_CSH, "CSh: About to close ccsSocket: %d", ccsSocket);	
        if (close(ccsSocket) == -1) 		/* Not needed by child */
            sysErrMsg("CSh: close ccsSocket");
	if (sigprocmask(SIG_SETMASK, &oldSet, NULL) == -1) 
	    sysErrMsg("CSh: sigprogmask SIG_SETMASK");
	
	dbgMsg(DBG_CSH, "CSh: About to close ptyMasterFd: %d", ptyMasterFd);	
        if (close(ptyMasterFd) == -1) 		/* Not needed by child */
            sysErrMsg("CSh: close ptyMasterFd");

        childShell(ccsInitSocket, ptyName, ccsSocketPort, pwd,
                authKey, clientProtocol, userAgent);
        exit(EXIT_FAILURE);		/* NOTREACHED */

    case -1:
        childShellAlive = FALSE;
    	sysErrMsg("startShellSession: fork");
        sendLoginFailureMsg(ccsInitSocket,
                "Child Command Server couldn't create Child Shell process");
        exit(EXIT_FAILURE);

    default: /* Parent */

	globChildShellPid = childShellPid;

        /* Create utmp and wtmp entries so child shell is visible 
	   to lastlog, w(1), etc */

	createUtmpAndWtmpEntries(childShellPid, pwd, ptyName);

	/*
	dbgMsg(DBG_CCS | DBG_CSH, "CCS (%d): successfully created CSh (%d)", 
		getpid(), childShellPid);
		*/
	/* Now it's okay to unblock SIGCHLD, so that if child has already 
	   died, now we'll catch SIGCHLD */

	if (sigprocmask(SIG_SETMASK, &oldSet, NULL) == -1) 
	    sysErrMsg("CCS: sigprogmask SIG_SETMASK");

	dbgMsg(DBG_CCS, "CCS: About to close ccsInitSocket: %d", ccsInitSocket);	
        if (close(ccsInitSocket) == -1)	/* Only childShell needs this now */
            sysErrMsg("startShellSession: close ccsInitSocket");

        childCommandServer(ccsSocket, ptyMasterFd, childShellPid, 
		authKey, pwd, clientProtocol);
        exit(EXIT_FAILURE);		/* NOTREACHED */

    } /* switch */
} /* startShellSession */



        /*********************************/
        /***** Master Command Server *****/
        /*********************************/

/* masterCommandServer()

   After opening a socket on the well known port specified in 'service'
   (which may either be a dotted IP address or a service name in
   /etc/services), we continuously loop, accepting new connections from
   the HTTP server, and starting up corresponding Client Shell sessions
   (a process pair consisting of a Child Command Server and a Child Shell).
*/

void
masterCommandServer(void)
{
    int mcsSocket;
    int ccsInitSocket;
    int addrLen;
    struct sockaddr_in clientAddr;
    pid_t ccsPid, dpid;

    
    /* If we were not started from inetd, then we must become a daemon
       and create a socket to accept future connection requests.  */
    
    if (runAsDaemon) {
    	dbgMsg(DBG_MCS, "Becoming a daemon");
	    
    	dpid = fork();
        if (dpid == -1) sysErrExit("fork - couldn't daemonise");
        if (dpid != 0) exit(EXIT_SUCCESS);	/* Parent exits */
        setsid();
        umask(0);
        chdir("/");
	weAreDaemon = TRUE;
    } /* if */

    /* Create listening socket for all incoming clients */

    dbgMsg(DBG_MCS, "Creating MCS socket");
    if ((mcsSocket = inetStreamSocket(service)) == -1)
    	    fatalErr("MCS: Couldn't create listening socket");
    
    if (listen(mcsSocket, 1) == -1) sysErrExit("MCS: listen");


    dbgMsg(DBG_MCS, "MCS PID=%d", getpid());

    if (signal(SIGCHLD, ccsReaper) == SIG_ERR)
    	sysErrExit("MCS: signal: ccsReaper");

    /* Process incoming connections, and start new shell sessions
       (= CCS + Child Shell) */

    for (;;) {

        /* Wait for next connection */

        addrLen = sizeof(clientAddr);
        memset(&clientAddr, 0, sizeof(struct sockaddr_in));
	dbgMsg(DBG_MCS, 
		"MCS about to wait for connection - socketfd = %d, issock=%d", 
		mcsSocket, isSocket(mcsSocket));
        ccsInitSocket = accept(mcsSocket, (struct sockaddr *) &clientAddr, 
		&addrLen);

        if (ccsInitSocket == -1) {
            if (errno == EINTR)		/* Interrupted by SIGCHLD handler */
            	continue;
            else 
            	sysErrExit("MCS: accept");
        } /* if */

        syslog(LOG_INFO, "Accepted connection from %s", 
		inetAddressString(&clientAddr)); 

        if (!isPermittedHttpServer(clientAddr.sin_addr.s_addr)) {
            syslog(LOG_NOTICE, "Denied access from %s",
            		inetAddressString(&clientAddr));
            sendLoginFailureMsg(ccsInitSocket, 
                    	"Access not permitted from this HTTP server");
            continue;	
        } /* if */

        /* All actual work for HTTP session is done by the CCS process */

        ccsPid = fork();

        if (ccsPid == 0) { 	/* CHILD */
	    dbgMsg(DBG_CCS, "CCS: About to close mcsSocket: %d", mcsSocket);	
            if (close(mcsSocket) == -1)   /* Child doesn't need this socket */
            	sysErrMsg("CCS: %d - close mcsSocket", getpid());

            startShellSession(ccsInitSocket);

            /* NOTREACHED */
            exit(EXIT_SUCCESS);
        } /* if */

        /* PARENT */

        if (ccsPid == -1) {
            sysErrMsg("MCS: fork");
            sendLoginFailureMsg(ccsInitSocket,
                        "Couldn't create Child Command Server process");
            continue;
        } /* if */

	dbgMsg(DBG_MCS, "MCS: About to close ccsInitSocket: %d", ccsInitSocket);
        if (close(ccsInitSocket) == -1) 	/* Parent doesn't need this */
            sysErrMsg("MCS: close ccsInitSocket");
    } /* for */
} /* masterCommandServer */



/* If we are started from inetd, then there is no MCS as such. 
   Instead inetd starts a new CCS for each connection.  Therefore we
   are already a daemon, and we already have a connected socket on
   stdin/stdout/stderr.  So we shortcut the MCS setup, do a little 
   initialisation and then start a CCS+CSh pair  */
	
void
inetdShellSession(void)
{
    int ccsInitSocket;
    int addrLen;
    struct sockaddr_in clientAddr;

    dbgMsg(DBG_MCS, "Stdin is socket: must have been started from inetd");

    ccsInitSocket = dup(STDIN_FILENO);

    if (close(STDIN_FILENO) == -1) 
	sysErrMsg("inetdShellSession: close STDIN");
    if (close(STDOUT_FILENO) == -1) 
	sysErrMsg("inetdShellSession: close STDOUT");
    if (close(STDERR_FILENO) == -1) 
	sysErrMsg("inetdShellSession: close STDERR");
	
    addrLen = sizeof(clientAddr);
    if (getsockname(ccsInitSocket, (struct sockaddr *) &clientAddr, 
		    &addrLen) == -1) 
	sysErrExit("getsockname");

    syslog(LOG_INFO, "Accepted connection from %s", 
		inetAddressString(&clientAddr)); 

    if (!isPermittedHttpServer(clientAddr.sin_addr.s_addr)) {
        syslog(LOG_NOTICE, "Denied access from %s",
            		inetAddressString(&clientAddr));
        sendLoginFailureMsg(ccsInitSocket, 
                    	"Access not permitted from this HTTP server");
    } /* if */

    startShellSession(ccsInitSocket);

    dbgMsg(DBG_MCS, "exiting inetdShellSession %s", getpid());
    exit(EXIT_SUCCESS);
} /* inetdShellSession */



	/************************/
	/***** Main program *****/
	/************************/

void
usage(void)
{
    if (!weAreDaemon) {
        fprintf(stderr, 
"htsh options are:\n"
"-b numbytes         Specify upper limit for per-user shell output buffers\n"
"                    (Default: %d bytes)\n"
"-c numbytes         Maximum number of bytes to be transferred at a time\n" 
"                    between Client Shell and HTTP Server.\n"
"                    (Default: %d bytes, minimum setting %d bytes)\n"
"-d                  Run in debug mode\n"
"-D [y|n]            If argument is 'y' run as daemon, otherwise ('n')\n"
"                    do not daemonise.\n"
"                    (If run from inetd, htsh is always a daemon.  If run from a\n"
"                    shell, htsh is a daemon by default)\n"
"-f config-file      Use named file for startup configuration, instead of %s\n"
"-F                  Do not use standard configuration file (%s)\n"
"-h                  Display this help message\n"
"-g global-init-file Use named file as login initialisation file for all user\n"
"                    logins.  (Default: %s)\n"
"-H http-server      Specify HTTP server from which connections will\n"
"                    be accepted.  (Can be specified multiple times)\n"
"-L log-file-dir     Record script(1) style log of each login session\n"
"                    in named directory\n"
"-o numsecs          Timeout for shell output (Default: %3.1f seconds,\n"
"                    minimum: %3.1f, maximum: %4.1f seconds)\n"
"-p port-num         Specify TCP port on which htsh should listen for\n"
"                    connections.  (Default is %s in /etc/services)\n"
"-P 'proto-name... ' Specify list of allowed login protocols\n"
"                    (can be '%s' and or '%s'\n"
"-t numsecs          Timeout for shell input (Default: %d seconds)\n"
"-u username         Specify user who is allowed to login to htsh\n"
"                    (Can be repeated multiple times)\n"
"\n"
"The -c, -D, -o, -p, and -t options override corresponding options in the\n"
"configuration file.  The -h and -u options are additive with corresponding\n"
"options in the configuration file\n",
		DEF_OUTPUT_BUFFER_LIMIT,
		DEF_CS_MAX_TRANSFER_SIZE, CS_MAX_TRANSFER_SIZE_LOW_BOUND,
		CONFIG_FILE, CONFIG_FILE,
		GLOBAL_RC_FILE,
		DEF_OUTPUT_TIMEOUT, MIN_OUTPUT_TIMEOUT, MAX_OUTPUT_TIMEOUT,
		SERVICE_NAME, PRNAME_WAP, PRNAME_HTTP, DEF_INPUT_TIMEOUT);
    } /* if */
} /* usage */



/* Add a named HTTP server to the list of HTTP Servers permitted to 
   login on this host.  Return TRUE if successful, or FALSE if the server 
   name is 'bad' (not a dotted IP address or a name in /etc/services) */

Boolean
addHttpServer(char *serverName)
{
    struct hostent *hostInfo;
		
    /* First try converting the host name as though it is an
               Internet dot notation address (a.b.c.d) */
             
    if ((httpServer[numHttpServers] = inet_addr(serverName)) == INADDR_NONE) {
         
        /* If that fails try looking up address by host name */
 
        if ((hostInfo = gethostbyname(serverName)) == NULL) 
	    return FALSE;

        memcpy(&(httpServer[numHttpServers]), hostInfo->h_addr, 
			hostInfo->h_length);
    } /* if */
    
    numHttpServers++;
    
    return TRUE;
} /* addHttpServer */



/* diagnose an error in the config file */

Boolean numConfigFileErrors; 	/* Total number of errors in config file */

void
configFileError(int lineNum, char *line, char *format, ...)
{
    va_list argList;
    
    va_start(argList, format);
    
    errorMsg("Configuration file error (line %d): %s", lineNum, line);
    rawErrorMsg(format, argList);
    
    numConfigFileErrors++;
    
    va_end(argList);
} /* configFileError */



/* Process config file */

void
processConfigFile(char *configFile)
{
    FILE *fp;
    int lineNum;
    int numLines;
    const int MAX_CF_ERRORS = 10;
    const int MAX_CF_LINE_LEN = 2000;
    char line[MAX_CF_LINE_LEN];
    char wLine[MAX_CF_LINE_LEN];
    char *statusMsg;
    int lineLen;
    char *firstToken, *nextToken;

    numConfigFileErrors = 0;

    fp = fopen(configFile, "r");

    if (fp == NULL) {
	errorMsg("Could not find configuration file: ignoring");
	return;
    } /* if */

    dbgMsg(DBG_CFG, "Processing config file: %s", configFile);

    for (lineNum = 1, numLines = 0; ; lineNum += numLines) {

	numLines = getLine(fp, MAX_CF_LINE_LEN, line, &lineLen, &statusMsg);
	
	if (numLines == 0) break;	/* EOF */

	if (numLines < 0) {
	    configFileError(lineNum, line, statusMsg);
	    break;
	} /* if */

	if (line[lineLen-1] == '\n') 	/* Strip trailing newline */
	    line[lineLen-1] = '\0';

	/*dbgMsg(DBG_CFG, "Config file %d: %s", lineNum, line);*/

	strcpy(wLine, line);
	firstToken = strtok(wLine, " \t\n");
	if (firstToken == NULL || firstToken[0] == '#') 
	    continue;

	nextToken = strtok(NULL, " \t\n");
	if (nextToken == NULL) {
	    configFileError(lineNum, line, "Missing argument");
	    continue;
	} /* if */

	if (strcmp(firstToken, "runasdaemon") == 0) {
	    if (strcmp(nextToken, "y") == 0)
		runAsDaemon = TRUE;
	    else if (strcmp(nextToken, "n") == 0)
		runAsDaemon = FALSE;
	    else 
	        configFileError(lineNum, line,
			"'runasdaemon' setting should be 'y' or 'n'");
	    
	} else if (strcmp(firstToken, "port") == 0) {
	    service = strdup(nextToken);

	} else if (strcmp(firstToken, "globalprofile") == 0) {
	    globalRcFile = strdup(nextToken);
	    
	} else if (strcmp(firstToken, "logfiledirectory") == 0) {
	    logFileDir = strdup(nextToken);
	    
	} else if (strcmp(firstToken, "csmaxtransfersize") == 0) {
	    if (!isIntNumber(nextToken, 10))
		configFileError(lineNum, line, "Integer argument expected: %s", 
			nextToken);
	    else 
		csMaxTransferSize = max(atoi(nextToken), 
			CS_MAX_TRANSFER_SIZE_LOW_BOUND);	

	} else if (strcmp(firstToken, "shelltimeout") == 0) {
	    if (!isIntNumber(nextToken, 10))
		configFileError(lineNum, line, "Integer argument expected: %s",
			nextToken);
	    else 
		shellTimeout = atoi(nextToken);
	    
	} else if (strcmp(firstToken, "outputbufferlimit") == 0) {
	    if (!isIntNumber(nextToken, 10))
		configFileError(lineNum, line, "Integer argument expected: %s",
			nextToken);
	    else 
		outputBufferLimit = atoi(nextToken);
	    
	} else if (strcmp(firstToken, "csoutputtimeout") == 0) {
	    if (!isFloatNumber(nextToken))
		configFileError(lineNum, line, "Numeric argument expected: %s",
			nextToken);
	    else 
		csOutputTimeout = atof(nextToken);
	    
	} else if (strcmp(firstToken, "allowedprotocols") == 0) {
	    char *statusMsg;
	    
	    allowedProtocols = parseProtocolList(&line[strlen(firstToken)+1], 
		    				&statusMsg);
	    if (allowedProtocols == -1)
		configFileError(lineNum, line, "%s", statusMsg);
	    
	} else if (strcmp(firstToken, "user") == 0) {
	    while (nextToken != NULL) {
                userSet[userSetSize] = strdup(nextToken);
                userSetSize++;
		nextToken = strtok(NULL, " \t\n");
	    } /* while */
	    
	} else if (strcmp(firstToken, "httpserver") == 0) {
	    while (nextToken != NULL) {
	        if (!addHttpServer(nextToken)) 
		    configFileError(lineNum, line, 
			    "Bad HTTP server name: %s\n", nextToken);
		nextToken = strtok(NULL, " \t\n");
	    } /* while */
	} else {
	    configFileError(lineNum, line, "Unknown command");
	} /* if */

	if (numConfigFileErrors >= MAX_CF_ERRORS) {
	    configFileError(lineNum, "", 
		    "Too many errors in configuration file: abandoning");
	} /* if */
	
    } /* for */

    fclose(fp);

    if (numConfigFileErrors > 0) 
	fatalErr("Terminating because of errors in configuration file");

} /* processConfigFile */



/* Process the config file and command line arguments.

   The -f and -F command line arguments control processing of the config file.
   Since the command line args must be processed after the config file,
   we must perform a bit of trickery.  We first process all cmd line
   args (in order to get and -f or -F options), 
   saving the values of any arguments whose values should overrride
   corresponding config file options, then process the config file,
   and reprocess the saved command line argument values.
*/
void
processConfigFileAndCmdLine(int argc, char *argv[])
{
    int opt;

    /* Saved values from command line args - NULL means that 
       this option did not appear on cmd line */
    
    char *cl_csMaxTransferSize = NULL;
    char *cl_csOutputTimeout = NULL;
    char *cl_outputBufferLimit = NULL;
    char *cl_service = NULL;
    char *cl_shellTimeout = NULL;
    char *cl_runAsDaemon = NULL;
    char *cl_globalRcFile = NULL;
    char *cl_allowedProtocols = NULL;
    char *cl_logFileDir = NULL;
    char *configFile = NULL;

    /* Set defaults */

    globalRcFile = GLOBAL_RC_FILE;
    ignoreStdConfigFile = FALSE;
    debugMask = 0;
    runAsDaemon = TRUE;
    userSetSize = 0;
    service = SERVICE_NAME;
    shellTimeout = DEF_INPUT_TIMEOUT;
    csOutputTimeout = DEF_OUTPUT_TIMEOUT;
    outputBufferLimit = DEF_OUTPUT_BUFFER_LIMIT;
    csMaxTransferSize = DEF_CS_MAX_TRANSFER_SIZE;

    /* Process command line arguments */
	
    while ((opt = getopt(argc, argv, "b:c:d:D:f:Fg:hH:L:o:p:P:t:u:")) != -1) {
        switch (opt) {

	case 'b':
            cl_outputBufferLimit = optarg;
	    if (!isIntNumber(optarg, 10)) {
            	errorMsg("Bad output buffer limit");
                exit(EXIT_BAD_CMD_LINE);
            } /* if */
            break;
	    
        case 'c':	/* Maximum output block size for transmission from
                           Child Shell to HTTP Server */
            cl_csMaxTransferSize = optarg;
            if (atoi(cl_csMaxTransferSize) <= 0) {
            	errorMsg("Bad shell output block size");
                exit(EXIT_BAD_CMD_LINE);
            } /* if */
            break;

        case 'd':	/* Debug */
	    if (!isIntNumber(optarg, 8)) {
            	errorMsg("Bad arg to -d option: %s", optarg);
                exit(EXIT_BAD_CMD_LINE);
            } /* if */
	    debugMask = strtol(optarg, NULL, 8);

            break;

        case 'D':	/* Run as daemon */
            cl_runAsDaemon = optarg;
	    if (strcmp(optarg, "y") != 0 && strcmp(optarg, "n") != 0) {
		errorMsg("Bad value for -D: %s", optarg);
		exit(EXIT_BAD_CMD_LINE);
	    } /* if */
		
            break;

	case 'f':	/* Specify alternate config file name */
	    configFile = optarg;
	    break;

	case 'F':	/* Don't process config file */
	    ignoreStdConfigFile = TRUE;
	    break;

	case 'g':	/* Specify alternate name for global login init file */
	    cl_globalRcFile = optarg;
	    break;

        case 'h':
            usage();
            exit(EXIT_SUCCESS);

        case 'H':	/* Specify an HTTP server from which we will accept
                           connections */
	    if (!addHttpServer(optarg)) {
		errorMsg("Bad HTTP server name: %s\n", optarg);
                exit(EXIT_BAD_CMD_LINE);
            } /* if */
            break;

	case 'L':	/* Log file directory */
	    cl_logFileDir = optarg;
	    break;
	    
        case 'o':	/* Timeout for shell output */
	    cl_csOutputTimeout = optarg;
            csOutputTimeout = atof(optarg);
            if (csOutputTimeout < MIN_OUTPUT_TIMEOUT || 
                    	csOutputTimeout > MAX_OUTPUT_TIMEOUT) {
            	errorMsg("Bad output timeout value: %s", optarg);
                exit(EXIT_BAD_CMD_LINE);
            } /* if */
            break;

        case 'p':	/* Specify port number other than the default */
            cl_service = optarg;
            break;
	    
    	case 'P':
	    cl_allowedProtocols = optarg;
	    break;

        case 't':	/* Timeout for shell session */
            cl_shellTimeout = optarg;
            if (atoi(cl_shellTimeout) <= 0) {
            	errorMsg("Bad input timeout value: %s", optarg);
                exit(EXIT_BAD_CMD_LINE);
            } /* if */
            break;

        case 'u':	/* Specify a user who is allowed to login */
            userSet[userSetSize] = optarg;
            userSetSize++;
            break;

        default:
            usage();
            exit(EXIT_BAD_CMD_LINE);
        } /* switch */
    } /* while */

    /* Further command line arg checking */
    
    if (configFile != NULL && ignoreStdConfigFile) {
	errorMsg("Cannot specify `-f' and `-F' options together");
	exit(EXIT_BAD_CMD_LINE);
    } /* if */

    if (optind < argc) {
	errorMsg("Unexpected argument: %s", argv[optind]);
	exit(EXIT_BAD_CMD_LINE);
    } /* if */

    dbgMsg(DBG_CFG, "Number of permitted users: %d", userSetSize);
    {
	int j;
	for (j = 0; j < userSetSize; j++)
	    dbgMsg(DBG_CFG, "%d: %s", j, userSet[j]);
    }
    dbgMsg(DBG_CFG, "Number of permitted HTTP servers: %d", numHttpServers);
    {
	int j;
	for (j = 0; j < numHttpServers; j++)
	    dbgMsg(DBG_CFG, "%d: %d", j, httpServer[j]);
    }
    
    /* Process config file */
    
    if (configFile == NULL) configFile = CONFIG_FILE;

    if (!ignoreStdConfigFile)
        processConfigFile(configFile);
    
    /* Now handle command line options which override whatever might have 
       been in config file */

    if (cl_csMaxTransferSize != NULL) 
	csMaxTransferSize = atoi(cl_csMaxTransferSize);
    
    if (cl_runAsDaemon != NULL) 
	runAsDaemon = cl_runAsDaemon[0] == 'y';
    
    if (cl_shellTimeout != NULL) 
	shellTimeout = atoi(cl_shellTimeout);
    
    if (cl_service != NULL) 
	service = cl_service;
    
    if (cl_csOutputTimeout != NULL) 
	csOutputTimeout = atof(cl_csOutputTimeout);
    
    if (cl_outputBufferLimit != NULL) 
	outputBufferLimit = atof(cl_outputBufferLimit);
    
    if (cl_globalRcFile != NULL) 
	globalRcFile = cl_globalRcFile;
    
    if (cl_logFileDir != NULL) 
	logFileDir = cl_logFileDir;
    
    if (cl_allowedProtocols != NULL) {
	char *statusMsg;
	
	allowedProtocols = parseProtocolList(cl_allowedProtocols, &statusMsg);
	if (allowedProtocols == -1) {
	    errorMsg("Error in -P option: %s", statusMsg);
	    exit(EXIT_BAD_CMD_LINE);
	} /* if */
    } /* if */
} /* processConfigFileAndCmdLine */



int
main(int argc, char *argv[])
{
    Boolean startedByInetd;		/* Were we started by inetd? */

    if (geteuid() != 0) {
    	errorMsg("Must be run under root");
        exit(EXIT_FAILURE);
    } /* if */

    startedByInetd = isSocket(STDIN_FILENO);

    weAreDaemon = startedByInetd;

    processConfigFileAndCmdLine(argc, argv);

    openlog(argv[0], LOG_PID, LOG_USER);
    
    if (startedByInetd) 
	inetdShellSession();
    else
        masterCommandServer();

    exit(EXIT_SUCCESS);
} /* main */
